# -*- test-case-name: twext.who.ldap.test.test_service -*-
##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from __future__ import print_function

"""
LDAP directory service implementation.
"""

from Queue import Queue, Empty
from threading import RLock
from uuid import UUID

import collections
import ldap.async

from twisted.python.constants import Names, NamedConstant
from twisted.internet.defer import succeed, inlineCallbacks, returnValue
from twisted.internet.threads import deferToThreadPool
from twisted.cred.credentials import IUsernamePassword
from twisted.python.threadpool import ThreadPool
from twisted.internet import reactor

from twext.python.log import Logger
from twext.python.types import MappingProxyType

from ..idirectory import (
    DirectoryServiceError, DirectoryAvailabilityError,
    FieldName as BaseFieldName, RecordType as BaseRecordType,
    IPlaintextPasswordVerifier, DirectoryConfigurationError
)
from ..directory import (
    DirectoryService as BaseDirectoryService,
    DirectoryRecord as BaseDirectoryRecord,
)
from ..expression import (
    MatchExpression, ExistsExpression, BooleanExpression,
    CompoundExpression, Operand, MatchType
)
from ..util import ConstantsContainer
from ._constants import LDAPAttribute, LDAPObjectClass
from ._util import (
    ldapQueryStringFromMatchExpression,
    ldapQueryStringFromCompoundExpression,
    ldapQueryStringFromBooleanExpression,
    ldapQueryStringFromExistsExpression,
)
from zope.interface import implementer



#
# Exceptions
#

class LDAPError(DirectoryServiceError):
    """
    LDAP error.
    """

    def __init__(self, message, ldapError=None):
        super(LDAPError, self).__init__(message)
        self.ldapError = ldapError



class LDAPConfigurationError(ValueError):
    """
    LDAP configuration error.
    """



class LDAPConnectionError(DirectoryAvailabilityError):
    """
    LDAP connection error.
    """

    def __init__(self, message, ldapError=None):
        super(LDAPConnectionError, self).__init__(message)
        self.ldapError = ldapError



class LDAPBindAuthError(LDAPConnectionError):
    """
    LDAP bind auth error.
    """



class LDAPQueryError(LDAPError):
    """
    LDAP query error.
    """



#
# Data type extensions
#

class FieldName(Names):
    dn = NamedConstant()
    dn.description = u"distinguished name"

    memberDNs = NamedConstant()
    memberDNs.description = u"member DNs"
    memberDNs.multiValue = True



#
# LDAP schema descriptions
#

class RecordTypeSchema(object):
    """
    Describes the LDAP schema for a record type.
    """
    def __init__(self, relativeDN, attributes):
        """
        @param relativeDN: The relative distinguished name for the record type.
            This is prepended to the service's base distinguished name when
            searching for records of this type.
        @type relativeDN: L{unicode}

        @param attributes: Attribute/value pairs that are expected for records
            of this type.
        @type attributes: iterable of sequences containing two L{unicode}s
        """
        self.relativeDN = relativeDN
        self.attributes = tuple(tuple(pair) for pair in attributes)



# We use strings (constant.value) instead of constants for the values in
# these mappings because it's meant to be configurable by application users,
# and user input forms such as config files aren't going to be able to use
# the constants.

# Maps field name -> LDAP attribute names

# NOTE: you must provide a mapping for uid

DEFAULT_FIELDNAME_ATTRIBUTE_MAP = MappingProxyType({
    # FieldName.dn: (LDAPAttribute.dn.value,),
    # BaseFieldName.uid: (LDAPAttribute.dn.value,),
    BaseFieldName.guid: (LDAPAttribute.generatedUUID.value,),
    BaseFieldName.shortNames: (LDAPAttribute.uid.value,),
    BaseFieldName.fullNames: (LDAPAttribute.cn.value,),
    BaseFieldName.emailAddresses: (LDAPAttribute.mail.value,),
    BaseFieldName.password: (LDAPAttribute.userPassword.value,),
})

# Information about record types
DEFAULT_RECORDTYPE_SCHEMAS = MappingProxyType({

    BaseRecordType.user: RecordTypeSchema(
        # ou=person
        relativeDN=u"ou={0}".format(LDAPObjectClass.person.value),

        # (objectClass=inetOrgPerson)
        attributes=(
            (
                LDAPAttribute.objectClass.value,
                LDAPObjectClass.inetOrgPerson.value,
            ),
        ),
    ),

    BaseRecordType.group: RecordTypeSchema(
        # ou=groupOfNames
        relativeDN=u"ou={0}".format(LDAPObjectClass.groupOfNames.value),

        # (objectClass=groupOfNames)
        attributes=(
            (
                LDAPAttribute.objectClass.value,
                LDAPObjectClass.groupOfNames.value,
            ),
        ),
    ),

})



#
# Directory Service
#

class DirectoryService(BaseDirectoryService):
    """
    LDAP directory service.
    """

    log = Logger()

    fieldName = ConstantsContainer((BaseFieldName, FieldName))

    recordType = ConstantsContainer((
        BaseRecordType.user, BaseRecordType.group,
    ))


    def __init__(
        self,
        url,
        baseDN,
        credentials=None,
        timeout=None,
        tlsCACertificateFile=None,
        tlsCACertificateDirectory=None,
        useTLS=False,
        fieldNameToAttributesMap=DEFAULT_FIELDNAME_ATTRIBUTE_MAP,
        recordTypeSchemas=DEFAULT_RECORDTYPE_SCHEMAS,
        extraFilters=None,
        ownThreadpool=True,
        threadPoolMax=10,
        connectionMax=10,
        _debug=False,
    ):
        """
        @param url: The URL of the LDAP server to connect to.
        @type url: L{unicode}

        @param baseDN: The base DN for queries.
        @type baseDN: L{unicode}

        @param credentials: The credentials to use to authenticate with the
            LDAP server.
        @type credentials: L{IUsernamePassword}

        @param timeout: A timeout, in seconds, for LDAP queries.
        @type timeout: number

        @param tlsCACertificateFile: ...
        @type tlsCACertificateFile: L{FilePath}

        @param tlsCACertificateDirectory: ...
        @type tlsCACertificateDirectory: L{FilePath}

        @param useTLS: Enable the use of TLS.
        @type useTLS: L{bool}

        @param fieldNameToAttributesMap: A mapping of field names to LDAP
            attribute names.
        @type fieldNameToAttributesMap: mapping with L{NamedConstant} keys and
            sequence of L{unicode} values

        @param recordTypeSchemas: Schema information for record types.
        @type recordTypeSchemas: mapping from L{NamedConstant} to
            L{RecordTypeSchema}

        @param extraFilters: A dict (keyed off recordType) of extra filter
            fragments to AND in to any generated queries.
        @type extraFilters: L{dicts} of L{unicode}

        """

        self.url = url
        self._baseDN = baseDN
        self._credentials = credentials
        self._timeout = timeout
        self._extraFilters = extraFilters

        if tlsCACertificateFile is None:
            self._tlsCACertificateFile = None
        else:
            self._tlsCACertificateFile = tlsCACertificateFile.path

        if tlsCACertificateDirectory is None:
            self._tlsCACertificateDirectory = None
        else:
            self._tlsCACertificateDirectory = tlsCACertificateDirectory.path

        self._useTLS = useTLS

        if _debug:
            self._debug = 255
        else:
            self._debug = None

        if self.fieldName.recordType in fieldNameToAttributesMap:
            raise TypeError("Record type field may not be mapped")

        if BaseFieldName.uid not in fieldNameToAttributesMap:
            raise DirectoryConfigurationError("Mapping for uid required")

        self._fieldNameToAttributesMap = fieldNameToAttributesMap

        self._attributeToFieldNameMap = {}
        for name, attributes in fieldNameToAttributesMap.iteritems():
            for attribute in attributes:
                if ":" in attribute:
                    attribute, ignored = attribute.split(":", 1)
                self._attributeToFieldNameMap.setdefault(
                    attribute, []
                ).append(name)

        self._recordTypeSchemas = recordTypeSchemas

        attributesToFetch = set()
        for attributes in fieldNameToAttributesMap.values():
            for attribute in attributes:
                if ":" in attribute:
                    attribute, ignored = attribute.split(":", 1)
                attributesToFetch.add(attribute.encode("utf-8"))
        self._attributesToFetch = list(attributesToFetch)

        # Threaded connection pool. The connection size limit here is the size for connections doing queries.
        # There will also be one-off connections for authentications which also run in their own threads. Thus
        # the threadpool max ought to be larger than the connection max to allow for both pooled query connections
        # and one-off auth-only connections.

        self.ownThreadpool = ownThreadpool
        if self.ownThreadpool:
            self.threadpool = ThreadPool(minthreads=1, maxthreads=threadPoolMax, name="LDAPDirectoryService")
        else:
            # Use the default threadpool but adjust its size to fit our needs
            self.threadpool = reactor.getThreadPool()
            self.threadpool.adjustPoolsize(
                max(threadPoolMax, self.threadpool.max)
            )
        self.connectionMax = connectionMax
        self.connectionCreateLock = RLock()
        self.connections = []
        self.connectionQueue = Queue()
        self.poolStats = collections.defaultdict(int)
        self.activeCount = 0

        reactor.callWhenRunning(self.start)
        reactor.addSystemEventTrigger('during', 'shutdown', self.stop)


    def start(self):
        """
        Start up this service. Initialize the threadpool (if we own it).
        """
        if self.ownThreadpool:
            self.threadpool.start()


    def stop(self):
        """
        Stop the service. Stop the threadpool if we own it and do other clean-up.
        """
        if self.ownThreadpool:
            self.threadpool.stop()

        # FIXME: we should probably also close the pool of active connections too


    @property
    def realmName(self):
        return u"{self.url}".format(self=self)


    class Connection(object):
        """
        ContextManager object for getting a connection from the pool. On exit the connection
        will be put back in the pool if no exception was raised. Otherwise, the connection will be
        removed from the active connection list, which will allow a new "clean" connection to
        be created later if needed.
        """

        def __init__(self, ds):
            self.ds = ds

        def __enter__(self):
            self.connection = self.ds._getConnection()
            return self.connection

        def __exit__(self, exc_type, exc_val, exc_tb):
            if exc_type is None:
                self.ds._returnConnection(self.connection)
                return True
            else:
                self.ds._failedConnection(self.connection)
                return False


    def _getConnection(self):
        """
        Get a connection from the connection pool. This will retrieve a connection from the connection
        pool L{Queue} object. If the L{Queue} is empty, it will check to see whether a new connection can
        be created (based on the connection limit), and if so create that and use it. If no new
        connections can be created, it will block on the L{Queue} until an existing, in-use, connection
        is put back.
        """
        try:
            connection = self.connectionQueue.get(block=False)
        except Empty:
            # Note we use a lock here to prevent a race condition in which multiple requests for a new connection
            # could succeed even though the connection counts starts out one less than the maximum. This can happen
            # because self._connect() can take a while.
            self.connectionCreateLock.acquire()
            if len(self.connections) < self.connectionMax:
                connection = self._connect()
                self.connections.append(connection)
                self.connectionCreateLock.release()
            else:
                self.connectionCreateLock.release()
                self.poolStats["connection-blocked"] += 1
                connection = self.connectionQueue.get()

        self.poolStats["connection-{}".format(self.connections.index(connection))] += 1
        self.activeCount += 1
        self.poolStats["connection-max"] = max(self.poolStats["connection-max"], self.activeCount)
        return connection


    def _returnConnection(self, connection):
        """
        A connection is no longer needed - return it to the pool.
        """
        self.activeCount -= 1
        self.connectionQueue.put(connection)


    def _failedConnection(self, connection):
        """
        A connection has failed - remove it from the list of active connections. A new
        one will be created if needed.
        """
        self.activeCount -= 1
        self.poolStats["connection-errors"] += 1
        self.connections.remove(connection)


    def _connect(self):
        """
        Connect to the directory server.
        This will always be called in a thread to prevent blocking.

        @returns: The connection object.
        @rtype: L{ldap.ldapobject.LDAPObject}

        @raises: L{LDAPConnectionError} if unable to connect.
        """

        # FIXME: ldap connection objects are not thread safe, so let's set up
        # a connection pool

        self.log.debug("Connecting to LDAP at {log_source.url}")
        connection = self._newConnection()

        if self._credentials is not None:
            if IUsernamePassword.providedBy(self._credentials):
                try:
                    connection.simple_bind_s(
                        self._credentials.username,
                        self._credentials.password,
                    )
                    self.log.debug(
                        "Bound to LDAP as {credentials.username}",
                        credentials=self._credentials
                    )
                except (
                    ldap.INVALID_CREDENTIALS, ldap.INVALID_DN_SYNTAX
                ) as e:
                    self.log.error(
                        "Unable to bind to LDAP as {credentials.username}",
                        credentials=self._credentials
                    )
                    raise LDAPBindAuthError(
                        self._credentials.username, e
                    )

            else:
                raise LDAPConnectionError(
                    "Unknown credentials type: {0}"
                    .format(self._credentials)
                )

        return connection


    def _newConnection(self):
        """
        Create a new LDAP connection and initialize and start TLS if required.
        This will always be called in a thread to prevent blocking.

        @returns: The connection object.
        @rtype: L{ldap.ldapobject.LDAPObject}

        @raises: L{LDAPConnectionError} if unable to connect.
        """
        connection = ldap.initialize(self.url)

        # FIXME: Use trace_file option to wire up debug logging when
        # Twisted adopts the new logging stuff.

        for option, value in (
            (ldap.OPT_TIMEOUT, self._timeout),
            (ldap.OPT_X_TLS_CACERTFILE, self._tlsCACertificateFile),
            (ldap.OPT_X_TLS_CACERTDIR, self._tlsCACertificateDirectory),
            (ldap.OPT_DEBUG_LEVEL, self._debug),
        ):
            if value is not None:
                connection.set_option(option, value)

        if self._useTLS:
            self.log.debug("Starting TLS for {log_source.url}")
            connection.start_tls_s()

        return connection


    def _authenticateUsernamePassword(self, dn, password):
        """
        Open a secondary connection to the LDAP server and try binding to it
        with the given credentials

        @returns: True if the password is correct, False otherwise
        @rtype: deferred C{bool}

        @raises: L{LDAPConnectionError} if unable to connect.
        """
        return deferToThreadPool(
            reactor, self.threadpool,
            self._authenticateUsernamePassword_inThread, dn, password
        )


    def _authenticateUsernamePassword_inThread(self, dn, password):
        """
        Open a secondary connection to the LDAP server and try binding to it
        with the given credentials.
        This method is always called in a thread.

        @returns: True if the password is correct, False otherwise
        @rtype: C{bool}

        @raises: L{LDAPConnectionError} if unable to connect.
        """
        self.log.debug("Authenticating {dn}", dn=dn)
        connection = self._newConnection()


        try:
            connection.simple_bind_s(dn, password)
            self.log.debug("Authenticated {dn}", dn=dn)
            return True
        except (
            ldap.INVALID_CREDENTIALS, ldap.INVALID_DN_SYNTAX
        ):
            self.log.debug("Unable to authenticate {dn}", dn=dn)
            return False
        finally:
            # TODO: should we explicitly "close" the connection in a finally
            # clause rather than just let it go out of scope and be garbage collected
            # at some indeterminate point in the future? Up side is that we won't hang
            # on to the connection or other resources for longer than needed. Down side
            # is we will take a little bit of extra time in this call to close it down.
            # If we do decide to "close" then we probably have to use one of the "unbind"
            # methods on the L{LDAPObject}.
            connection = None


    def _recordsFromQueryString(
        self, queryString, recordTypes=None,
        limitResults=None, timeoutSeconds=None
    ):
        return deferToThreadPool(
            reactor, self.threadpool,
            self._recordsFromQueryString_inThread,
            queryString,
            recordTypes,
            limitResults=limitResults,
            timeoutSeconds=timeoutSeconds
        )


    def _addExtraFilter(self, recordType, queryString):
        if self._extraFilters and self._extraFilters.get(recordType, ""):
            queryString = "(&{extra}{query})".format(
                extra=self._extraFilters[recordType], query=queryString
            )
        return queryString


    def _recordsFromQueryString_inThread(
        self, queryString, recordTypes=None,
        limitResults=None, timeoutSeconds=None
    ):
        """
        This method is always called in a thread.
        """

        records = []

        with DirectoryService.Connection(self) as connection:


            if recordTypes is None:
                recordTypes = self.recordTypes()

            for recordType in recordTypes:

                if limitResults is not None:
                    if limitResults < 1:
                        break

                try:
                    rdn = self._recordTypeSchemas[recordType].relativeDN
                except KeyError:
                    # Skip this unknown record type
                    continue

                rdn = (
                    ldap.dn.str2dn(rdn.lower()) +
                    ldap.dn.str2dn(self._baseDN.lower())
                )
                filteredQuery = self._addExtraFilter(recordType, queryString)
                self.log.debug(
                    "Performing LDAP query: {rdn} {query} {recordType}{limit}{timeout}",
                    rdn=rdn,
                    query=filteredQuery,
                    recordType=recordType,
                    limit=" limit={}".format(limitResults) if limitResults else "",
                    timeout=" timeout={}".format(timeoutSeconds) if timeoutSeconds else "",
                )
                try:
                    s = ldap.async.List(connection)
                    s.startSearch(
                        ldap.dn.dn2str(rdn),
                        ldap.SCOPE_SUBTREE,
                        filteredQuery,
                        attrList=self._attributesToFetch,
                        timeout=timeoutSeconds if timeoutSeconds else -1,
                        sizelimit=limitResults if limitResults else 0
                    )
                    s.processResults()

                except ldap.SIZELIMIT_EXCEEDED as e:
                    self.log.debug("LDAP result limit exceeded: {}".format(limitResults,))

                except ldap.TIMELIMIT_EXCEEDED as e:
                    self.log.warn("LDAP timeout exceeded: {} seconds".format(timeoutSeconds,))

                except ldap.FILTER_ERROR as e:
                    self.log.error(
                        "Unable to perform query {query!r}: {err}",
                        query=queryString, err=e
                    )
                    raise LDAPQueryError("Unable to perform query", e)

                except ldap.NO_SUCH_OBJECT as e:
                    # self.log.warn("RDN {rdn} does not exist, skipping", rdn=rdn)
                    continue

                except ldap.INVALID_SYNTAX as e:
                    self.log.error(
                        "LDAP invalid syntax {query!r}: {err}",
                        query=queryString, err=e
                    )
                    continue

                except ldap.SERVER_DOWN as e:
                    self.log.error(
                        "LDAP server unavailable"
                    )
                    raise LDAPQueryError("LDAP server down", e)

                except Exception as e:
                    self.log.error(
                        "LDAP error {query!r}: {err}",
                        query=queryString, err=e
                    )
                    raise LDAPQueryError("Unable to perform query", e)

                reply = [resultItem for _ignore_resultType, resultItem in s.allResults]

                newRecords = self._recordsFromReply(reply, recordType=recordType)

                self.log.debug(
                    "Records from LDAP query ({rdn} {query} {recordType}): {count}",
                    rdn=rdn,
                    query=queryString,
                    recordType=recordType,
                    count=len(newRecords)
                )

                if limitResults is not None:
                    limitResults = limitResults - len(newRecords)

                records.extend(newRecords)

        self.log.debug(
            "LDAP result count ({query}): {count}",
            query=queryString,
            count=len(records)
        )

        return records


    def _recordWithDN(self, dn):
        return deferToThreadPool(
            reactor, self.threadpool,
            self._recordWithDN_inThread, dn
        )


    def _recordWithDN_inThread(self, dn):
        """
        This method is always called in a thread.

        @param dn: The DN of the record to search for
        @type dn: C{str}
        """
        with DirectoryService.Connection(self) as connection:

            self.log.debug("Performing LDAP DN query: {dn}", dn=dn)

            try:
                reply = connection.search_s(
                    dn,
                    ldap.SCOPE_SUBTREE,
                    "(objectClass=*)",
                    attrlist=self._attributesToFetch
                )
                records = self._recordsFromReply(reply)
            except ldap.NO_SUCH_OBJECT:
                records = []
            except ldap.INVALID_DN_SYNTAX:
                self.log.warn("Invalid LDAP DN syntax: '{dn}'", dn=dn)
                records = []

        if len(records):
            return records[0]
        else:
            return None


    def _recordsFromReply(self, reply, recordType=None):
        records = []

        for dn, recordData in reply:

            # Determine the record type
            if recordType is None:
                recordType = recordTypeForDN(
                    self._baseDN, self._recordTypeSchemas, dn
                )

            if recordType is None:
                recordType = recordTypeForRecordData(
                    self._recordTypeSchemas, recordData
                )

            if recordType is None:
                self.log.debug(
                    "Ignoring LDAP record data; unable to determine record "
                    "type: {recordData!r}",
                    recordData=recordData,
                )
                continue

            # Populate a fields dictionary
            fields = {}

            for fieldName, attributeRules in self._fieldNameToAttributesMap.iteritems():
                valueType = self.fieldName.valueType(fieldName)

                for attributeRule in attributeRules:
                    attributeName = attributeRule.split(":")[0]
                    if attributeName in recordData:
                        values = recordData[attributeName]

                        if valueType in (unicode, UUID):
                            if not isinstance(values, list):
                                values = [values]

                            if valueType is unicode:
                                newValues = []
                                for v in values:
                                    if isinstance(v, unicode):
                                        # because the ldap unit test produces
                                        # unicode values (?)
                                        newValues.append(v)
                                    else:
                                        newValues.append(unicode(v, "utf-8"))
                            else:
                                try:
                                    newValues = [valueType(v) for v in values]
                                except Exception, e:
                                    self.log.warn(
                                        "Can't parse value {name} {values} ({error})",
                                        name=fieldName, values=values, error=str(e)
                                    )
                                    continue

                            if self.fieldName.isMultiValue(fieldName):
                                if fieldName in fields:
                                    fields[fieldName].extend(newValues)
                                else:
                                    fields[fieldName] = newValues
                            else:
                                # First one in the list wins
                                if fieldName not in fields:
                                    fields[fieldName] = newValues[0]

                        elif valueType is bool:
                            if not isinstance(values, list):
                                values = [values]
                            if ":" in attributeRule:
                                ignored, trueValue = attributeRule.split(":")
                            else:
                                trueValue = "true"

                            for value in values:
                                if value == trueValue:
                                    fields[fieldName] = True
                                    break
                            else:
                                fields[fieldName] = False

                        elif issubclass(valueType, Names):
                            if not isinstance(values, list):
                                values = [values]

                            attribute, attributeValue, fieldValue = attributeRule.split(":")

                            for value in values:
                                if value == attributeValue:
                                    # convert to a constant
                                    try:
                                        fieldValue = valueType.lookupByName(fieldValue)
                                        fields[fieldName] = fieldValue
                                    except ValueError:
                                        pass
                                    break

                        else:
                            raise LDAPConfigurationError(
                                "Unknown value type {0} for field {1}".format(
                                    valueType, fieldName
                                )
                            )

            # Skip any results missing the uid, which is a required field
            if self.fieldName.uid not in fields:
                continue

            # Set record type and dn fields
            fields[self.fieldName.recordType] = recordType
            fields[self.fieldName.dn] = dn.decode("utf-8")

            # Make a record object from fields.
            record = DirectoryRecord(self, fields)
            records.append(record)

        # self.log.debug("LDAP results: {records}", records=records)

        return records


    def recordsFromNonCompoundExpression(
        self, expression, recordTypes=None, records=None, limitResults=None,
        timeoutSeconds=None
    ):
        if isinstance(expression, MatchExpression):
            queryString = ldapQueryStringFromMatchExpression(
                expression,
                self._fieldNameToAttributesMap, self._recordTypeSchemas
            )
            return self._recordsFromQueryString(
                queryString, recordTypes=recordTypes,
                limitResults=limitResults, timeoutSeconds=timeoutSeconds
            )

        elif isinstance(expression, ExistsExpression):
            queryString = ldapQueryStringFromExistsExpression(
                expression,
                self._fieldNameToAttributesMap, self._recordTypeSchemas
            )
            return self._recordsFromQueryString(
                queryString, recordTypes=recordTypes,
                limitResults=limitResults, timeoutSeconds=timeoutSeconds
            )

        elif isinstance(expression, BooleanExpression):
            queryString = ldapQueryStringFromBooleanExpression(
                expression,
                self._fieldNameToAttributesMap, self._recordTypeSchemas
            )
            return self._recordsFromQueryString(
                queryString, recordTypes=recordTypes,
                limitResults=limitResults, timeoutSeconds=timeoutSeconds
            )

        return BaseDirectoryService.recordsFromNonCompoundExpression(
            self, expression, records=records, limitResults=limitResults,
            timeoutSeconds=timeoutSeconds
        )


    def recordsFromCompoundExpression(
        self, expression, recordTypes=None, records=None,
        limitResults=None, timeoutSeconds=None
    ):
        if not expression.expressions:
            return succeed(())

        queryString = ldapQueryStringFromCompoundExpression(
            expression,
            self._fieldNameToAttributesMap, self._recordTypeSchemas
        )
        return self._recordsFromQueryString(
            queryString, recordTypes=recordTypes,
            limitResults=limitResults, timeoutSeconds=timeoutSeconds
        )


    def recordsWithRecordType(
        self, recordType, limitResults=None, timeoutSeconds=None
    ):
        queryString = ldapQueryStringFromExistsExpression(
            ExistsExpression(self.fieldName.uid),
            self._fieldNameToAttributesMap, self._recordTypeSchemas
        )
        return self._recordsFromQueryString(
            queryString, recordTypes=[recordType],
            limitResults=limitResults, timeoutSeconds=timeoutSeconds
        )


    # def updateRecords(self, records, create=False):
    #     for record in records:
    #         return fail(NotAllowedError("Record updates not allowed."))
    #     return succeed(None)


    # def removeRecords(self, uids):
    #     for uid in uids:
    #         return fail(NotAllowedError("Record removal not allowed."))
    #     return succeed(None)



@implementer(IPlaintextPasswordVerifier)
class DirectoryRecord(BaseDirectoryRecord):
    """
    LDAP directory record.
    """

    @inlineCallbacks
    def members(self):

        members = set()

        if self.recordType != self.service.recordType.group:
            returnValue(())

        # Scan through the memberDNs, grouping them by record type (which we
        # deduce by their RDN).  If we have a fieldname that corresponds to
        # the most specific slice of the DN, we can bundle that into a
        # single CompoundExpression to fault in all the DNs belonging to the
        # same base RDN, reducing the number of requests from 1-per-member to
        # 1-per-record-type.  Any memberDNs we can't group in this way are
        # simply faulted in by DN at the end.

        fieldValuesByRecordType = {}
        # dictionary key = recordType, value = tuple(fieldName, value)

        faultByDN = []
        # the DNs we need to fault in individually

        for dnStr in getattr(self, "memberDNs", []):
            try:
                recordType = recordTypeForDN(
                    self.service._baseDN, self.service._recordTypeSchemas, dnStr
                )
                dn = ldap.dn.str2dn(dnStr.lower())
                attrName, value, ignored = dn[0][0]
                fieldName = self.service._attributeToFieldNameMap[attrName][0]
                fieldValuesByRecordType.setdefault(recordType, []).append((fieldName, value))
                continue

            except:
                # For whatever reason we can't group this DN in with the others
                # so we'll add it to faultByDN just below
                pass

            # have to fault in by dn
            faultByDN.append(dnStr)

        for recordType, fieldValue in fieldValuesByRecordType.iteritems():
            if fieldValue:
                matchExpressions = []
                for fieldName, value in fieldValue:
                    matchExpressions.append(
                        MatchExpression(
                            fieldName,
                            value.decode("utf-8"),
                            matchType=MatchType.equals
                        )
                    )
            expression = CompoundExpression(
                matchExpressions,
                Operand.OR
            )
            for record in (yield self.service.recordsFromCompoundExpression(
                expression, recordTypes=[recordType]
            )):
                members.add(record)

        for dnStr in faultByDN:
            record = yield self.service._recordWithDN(dnStr.replace("+", "\+"))
            members.add(record)

        returnValue(members)


    # @inlineCallbacks
    def groups(self):
        raise NotImplementedError()


    #
    # Verifiers for twext.who.checker stuff.
    #

    def verifyPlaintextPassword(self, password):
        return self.service._authenticateUsernamePassword(self.dn, password)



def normalizeDNstr(dnStr):
    """
    Convert to lowercase and remove extra whitespace
    @param dnStr: dn
    @type dnStr: C{str}
    @return: normalized dn C{str}
    """
    return ' '.join(ldap.dn.dn2str(ldap.dn.str2dn(dnStr.lower())).split())



def reverseDict(source):
    new = {}

    for key, values in source.iteritems():
        for value in values:
            new.setdefault(value, []).append(key)

    return new



def recordTypeForDN(baseDnStr, recordTypeSchemas, dnStr):
    """
    Examine a DN to determine which recordType it belongs to

    @param baseDnStr: The base DN
    @type baseDnStr: string

    @param recordTypeSchemas: Schema information for record types.
    @type recordTypeSchemas: mapping from L{NamedConstant} to
        L{RecordTypeSchema}

    @param dnStr: DN to compare
    @type dnStr: string

    @return: recordType string, or None if no match
    """
    dn = ldap.dn.str2dn(dnStr.lower())
    baseDN = ldap.dn.str2dn(baseDnStr.lower())

    for recordType, schema in recordTypeSchemas.iteritems():
        combined = ldap.dn.str2dn(schema.relativeDN.lower()) + baseDN
        if dnContainedIn(dn, combined):
            return recordType
    return None



def dnContainedIn(child, parent):
    """
    Return True if child dn is contained within parent dn, otherwise False.
    """
    return child[-len(parent):] == parent



def recordTypeForRecordData(recordTypeSchemas, recordData):
    """
    Given info about record types, determine the record type for a blob of
    LDAP record data.

    @param recordTypeSchemas: Schema information for record types.
    @type recordTypeSchemas: mapping from L{NamedConstant} to
        L{RecordTypeSchema}

    @param recordData: LDAP record data.
    @type recordData: mapping
    """

    for recordType, schema in recordTypeSchemas.iteritems():
        for attribute, value in schema.attributes:
            dataValue = recordData.get(attribute)
            # If the data value (e.g. objectClass) is a list, see if the
            # expected value is contained in that list, otherwise directly
            # compare.
            if isinstance(dataValue, list):
                if value not in dataValue:
                    break
            else:
                if value != dataValue:
                    break
        else:
            return recordType

    return None



# -*- test-case-name: twext.who.test.test_directory -*-
##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Generic directory service base implementation
"""

__all__ = [
    "DirectoryService",
    "DirectoryRecord",
]

from zope.interface import implementer, directlyProvides

from twisted.python.constants import (
    Names, NamedConstant,
    Values, ValueConstant,
    Flags, FlagConstant,
)
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.defer import succeed, fail
from twisted.cred.credentials import DigestedCredentials

from .idirectory import (
    InvalidDirectoryRecordError,
    QueryNotSupportedError, NotAllowedError,
    FieldName, RecordType,
    IDirectoryService, IDirectoryRecord,
    IPlaintextPasswordVerifier, IHTTPDigestVerifier,
)
from .expression import CompoundExpression, Operand, MatchExpression
from .util import uniqueResult, describe, ConstantsContainer



@implementer(IDirectoryService)
class DirectoryService(object):
    """
    Generic (and abstract) implementation of L{IDirectoryService}.

    Most of the C{recordsWith*} methods call L{recordsWithFieldValue}, which in
    turn calls L{recordsFromExpression} with a corresponding
    L{MatchExpression}.

    L{recordsFromExpression} relies on L{recordsFromNonCompoundExpression} for
    all expression types other than L{CompoundExpression}, which it handles
    directly.

    L{recordsFromNonCompoundExpression} (and therefore most uses of the other
    methods) will always fail with a L{QueryNotSupportedError}.

    A subclass should therefore override L{recordsFromNonCompoundExpression}
    with an implementation that handles any queries that it can support (which
    should include L{MatchExpression}) and calls its superclass' implementation
    with any query it cannot support.

    A subclass may override L{recordsFromExpression} if it is to support
    L{CompoundExpression}s with operands other than L{Operand.AND} and
    L{Operand.OR}.

    A subclass may override L{recordsFromExpression} if it is built on top
    of a directory service that supports compound expressions, as that may be
    more effient than relying on L{DirectoryService}'s implementation.

    L{updateRecords} and L{removeRecords} will fail with L{NotAllowedError}
    when asked to modify data.
    A subclass should override these methods if is to allow editing of
    directory information.

    @cvar recordType: a L{Names} class or compatible object (eg.
        L{ConstantsContainer}) which contains the L{NamedConstant}s denoting
        the record types that are supported by this directory service.

    @cvar fieldName: a L{Names} class or compatible object (eg.
        L{ConstantsContainer}) which contains the L{NamedConstant}s denoting
        the record field names that are supported by this directory service.

    @cvar normalizedFields: a L{dict} mapping of (ie. L{NamedConstant}s
        contained in the C{fieldName} class variable) to callables that take
        a field value (a L{unicode}) and return a normalized field value (also
        a L{unicode}).
    """

    recordType = ConstantsContainer(())
    fieldName = FieldName

    normalizedFields = {
        FieldName.emailAddresses: lambda e: e.lower(),
    }


    def __init__(self, realmName):
        """
        @param realmName: a realm name
        @type realmName: L{unicode}
        """
        self.realmName = realmName


    def __repr__(self):
        return (
            "<{self.__class__.__name__} {self.realmName!r}>"
            .format(self=self)
        )


    def recordTypes(self):
        return self.recordType.iterconstants()


    def recordsFromNonCompoundExpression(
        self, expression, recordTypes=None, records=None,
        limitResults=None, timeoutSeconds=None
    ):
        """
        Finds records matching a non-compound expression.

        @note: This method is called by L{recordsFromExpression} to handle
            all expressions other than L{CompoundExpression}.
            This implementation always fails with L{QueryNotSupportedError}.
            Subclasses should override this in order to handle additional
            expression types, and call on the superclass' implementation
            for other expression types.

        @note: This interface is the same as L{recordsFromExpression}, except
            for the additional C{records} argument.

        @param expression: an expression to apply
        @type expression: L{object}

        @param recordTypes: the record types to match
        @type recordTypes: an iterable of L{NamedConstant}, or None for no
            filtering

        @param records: a set of records to limit the search to. C{None} if
            the whole directory should be searched.
            This is provided by L{recordsFromExpression} when it has already
            narrowed down results to a set of records.
            That is, it's a performance optimization; ignoring this and
            searching the entire directory will also work.
        @type records: L{set} or L{frozenset}

        @return: The matching records.
        @rtype: deferred iterable of L{IDirectoryRecord}s

        @raises: L{QueryNotSupportedError} if the expression is not
            supported by this directory service.
        """
        if records is not None:
            for _ignore_record in records:
                break
            else:
                return succeed(())

        return fail(QueryNotSupportedError(
            "Unknown expression: {0}".format(expression)
        ))


    @inlineCallbacks
    def recordsFromCompoundExpression(
        self, expression, recordTypes=None, records=None,
        limitResults=None, timeoutSeconds=None
    ):
        """
        Finds records matching a compound expression.

        @note: This method is called by L{recordsFromExpression} to handle
            all L{CompoundExpression}s.

        @note: This interface is the same as L{recordsFromExpression}, except
            for the additional C{records} argument.

        @param expression: an expression to apply
        @type expression: L{CompoundExpression}

        @param recordTypes: the record types to match
        @type recordTypes: an iterable of L{NamedConstant}, or None for no
            filtering

        @param records: a set of records to limit the search to. C{None} if
            the whole directory should be searched.
            This is provided by L{recordsFromExpression} when it has already
            narrowed down results to a set of records.
            That is, it's a performance optimization; ignoring this and
            searching the entire directory will also work.
        @type records: L{set} or L{frozenset}

        @return: The matching records.
        @rtype: deferred iterable of L{IDirectoryRecord}s

        @raises: L{QueryNotSupportedError} if the expression is not
            supported by this directory service.
        """
        operand = expression.operand
        subExpressions = iter(expression.expressions)

        try:
            subExpression = subExpressions.next()
        except StopIteration:
            returnValue(())

        results = set((
            yield self.recordsFromExpression(
                subExpression, recordTypes=recordTypes, records=records
            )
        ))

        for subExpression in subExpressions:
            if operand == Operand.AND:
                if not results:
                    # No need to bother continuing here
                    returnValue(())

                records = results
            else:
                records = None

            recordsMatchingExpression = frozenset((
                yield self.recordsFromExpression(
                    subExpression, recordTypes=recordTypes, records=records
                )
            ))

            if operand == Operand.AND:
                results &= recordsMatchingExpression
            elif operand == Operand.OR:
                results |= recordsMatchingExpression
            else:
                raise QueryNotSupportedError(
                    "Unknown operand: {0}".format(operand)
                )

        returnValue(results)


    def recordsFromExpression(
        self, expression, recordTypes=None, records=None,
        limitResults=None, timeoutSeconds=None
    ):
        """
        @note: This interface is the same as
            L{IDirectoryService.recordsFromExpression}, except for the
            additional C{records} argument.
        """
        if isinstance(expression, CompoundExpression):
            return self.recordsFromCompoundExpression(
                expression, recordTypes=recordTypes,
                limitResults=limitResults, timeoutSeconds=timeoutSeconds
            )
        else:
            return self.recordsFromNonCompoundExpression(
                expression, recordTypes=recordTypes,
                limitResults=limitResults, timeoutSeconds=timeoutSeconds
            )


    def recordsWithFieldValue(
        self, fieldName, value, recordTypes=None,
        limitResults=None, timeoutSeconds=None
    ):
        return self.recordsFromExpression(
            MatchExpression(fieldName, value),
            recordTypes=recordTypes,
            limitResults=limitResults,
            timeoutSeconds=timeoutSeconds
        )


    @inlineCallbacks
    def recordWithUID(self, uid, timeoutSeconds=None):
        returnValue(uniqueResult(
            (yield self.recordsWithFieldValue(
                FieldName.uid, uid, timeoutSeconds=timeoutSeconds
            ))
        ))


    @inlineCallbacks
    def recordWithGUID(self, guid, timeoutSeconds=None):
        returnValue(uniqueResult(
            (yield self.recordsWithFieldValue(
                FieldName.guid, guid, timeoutSeconds=timeoutSeconds
            ))
        ))


    def recordsWithRecordType(
        self, recordType, limitResults=None, timeoutSeconds=None
    ):
        return self.recordsWithFieldValue(
            FieldName.recordType, recordType,
            limitResults=limitResults, timeoutSeconds=timeoutSeconds
        )


    @inlineCallbacks
    def recordWithShortName(self, recordType, shortName, timeoutSeconds=None):
        returnValue(
            uniqueResult(
                (
                    yield self.recordsFromExpression(
                        MatchExpression(
                            FieldName.shortNames, shortName
                        ),
                        recordTypes=[recordType],
                        timeoutSeconds=timeoutSeconds
                    )
                )
            )
        )


    def recordsWithEmailAddress(
        self, emailAddress, limitResults=None, timeoutSeconds=None
    ):
        return self.recordsWithFieldValue(
            FieldName.emailAddresses,
            emailAddress,
            limitResults=limitResults,
            timeoutSeconds=timeoutSeconds
        )


    def updateRecords(self, records, create=False):
        for _ignore_record in records:
            return fail(NotAllowedError("Record updates not allowed."))
        return succeed(None)


    def removeRecords(self, uids):
        for _ignore_uid in uids:
            return fail(NotAllowedError("Record removal not allowed."))
        return succeed(None)


    def flush(self):
        return succeed(None)



@implementer(IDirectoryRecord)
class DirectoryRecord(object):
    """
    Generic implementation of L{IDirectoryService}.

    This is an incomplete implementation of L{IDirectoryRecord}.

    L{groups} will always fail with L{NotImplementedError} and L{members} will
    do so if this is a group record.
    A subclass should override these methods to support group membership and
    complete this implementation.

    @cvar requiredFields: an iterable of field names that must be present in
        all directory records.
    """

    requiredFields = (
        FieldName.uid,
        FieldName.recordType,
    )


    def __init__(self, service, fields):
        for fieldName in self.requiredFields:
            if fieldName not in fields or not fields[fieldName]:
                raise InvalidDirectoryRecordError(
                    "{0} field is required.".format(fieldName),
                    fields
                )

            if service.fieldName.isMultiValue(fieldName):
                values = fields[fieldName]
                for value in values:
                    if not value:
                        raise InvalidDirectoryRecordError(
                            "{0} field must not be empty.".format(fieldName),
                            fields
                        )

        if (
            fields[FieldName.recordType] not in
            service.recordType.iterconstants()
        ):
            raise InvalidDirectoryRecordError(
                "Unknown record type: {0!r} is not in {1!r}.".format(
                    fields[FieldName.recordType],
                    tuple(service.recordType.iterconstants()),
                ),
                fields
            )

        def checkType(name, value):
            expectedType = service.fieldName.valueType(name)

            if issubclass(expectedType, Flags):
                expectedType = FlagConstant
            elif issubclass(expectedType, Values):
                expectedType = ValueConstant
            elif issubclass(expectedType, Names):
                expectedType = NamedConstant

            if value is not None and not isinstance(value, expectedType):
                raise InvalidDirectoryRecordError(
                    "Value {0!r} for field {1} is not of type {2}".format(
                        value, name, expectedType
                    ),
                    fields
                )

        # Normalize fields
        normalizedFields = {}
        for name, value in fields.items():
            if not isinstance(name, NamedConstant):
                raise InvalidDirectoryRecordError(
                    "Field name {} is not a named constant".format(name),
                    fields
                )

            normalize = service.normalizedFields.get(name, None)

            if normalize is None:
                normalizedValue = value
            else:
                if service.fieldName.isMultiValue(name):
                    normalizedValue = tuple((normalize(v) for v in value))
                else:
                    normalizedValue = normalize(value)

            if service.fieldName.isMultiValue(name):
                for v in normalizedValue:
                    checkType(name, v)
            else:
                checkType(name, normalizedValue)

            normalizedFields[name] = normalizedValue

        self.service = service
        self.fields = normalizedFields

        if self.service.fieldName.password in self.fields:
            directlyProvides(
                self, IPlaintextPasswordVerifier, IHTTPDigestVerifier
            )


    def __repr__(self):
        try:
            return (
                "<{self.__class__.__name__} ({recordType}){shortName}>".format(
                    self=self,
                    recordType=describe(self.recordType),
                    shortName=self.shortNames[0],
                )
            )
        except (AttributeError, IndexError):
            return (
                "<{self.__class__.__name__} {uid}>".format(
                    self=self,
                    uid=self.uid,
                )
            )


    def __hash__(self):
        return hash(self.uid)


    def __eq__(self, other):
        if IDirectoryRecord.implementedBy(other.__class__):
            return (
                self.service == other.service and
                self.fields == other.fields
            )
        return NotImplemented


    def __ne__(self, other):
        eq = self.__eq__(other)
        if eq is NotImplemented:
            return NotImplemented
        return not eq


    def __getattr__(self, name):
        try:
            fieldName = self.service.fieldName.lookupByName(name)
        except ValueError:
            raise AttributeError(name)

        try:
            return self.fields[fieldName]
        except KeyError:
            raise AttributeError(name)


    def description(self):
        """
        Generate a string description of this directory record.

        @return: A description.
        @rtype: L{unicode}
        """
        def describeValue(value):
            if hasattr(value, "description"):
                value = value.description
            else:
                value = unicode(value)

            return value

        values = {}

        for fieldName in self.service.fieldName.iterconstants():
            if fieldName not in self.fields:
                continue

            value = self.fields[fieldName]

            if hasattr(fieldName, "description"):
                name = fieldName.description
            else:
                name = unicode(fieldName.name)

            if self.service.fieldName.isMultiValue(fieldName):
                values[name] = ", ".join(describeValue(v) for v in value)
            else:
                values[name] = describeValue(value)

        description = [self.__class__.__name__, u":"]

        for name in sorted(values.iterkeys()):
            description.append(u"\n  {0} = {1}".format(name, values[name]))

        description.append(u"\n")

        return u"".join(description)


    def members(self):
        if self.recordType == RecordType.group:
            return fail(
                NotImplementedError("Subclasses must implement members()")
            )
        return succeed(())


    def groups(self):
        return fail(NotImplementedError("Subclasses must implement groups()"))


    def addMembers(self, members):
        return fail(NotAllowedError("Membership updates not allowed."))


    def removeMembers(self, members):
        return fail(NotAllowedError("Membership updates not allowed."))


    def setMembers(self, members):
        return fail(NotAllowedError("Membership updates not allowed."))


    #
    # Verifiers for twext.who.checker stuff.
    #

    def verifyPlaintextPassword(self, password):
        if self.password == password:
            return succeed(True)
        else:
            return succeed(False)


    def verifyHTTPDigest(
        self, username, realm, uri, nonce, cnonce,
        algorithm, nc, qop, response, method,
    ):
        helperCreds = DigestedCredentials(
            username, method, realm,
            dict(
                realm=realm,
                uri=uri,
                nonce=nonce,
                cnonce=cnonce,
                algorithm=algorithm,
                nc=nc,
                qop=qop,
                response=response
            )
        )

        return succeed(helperCreds.checkPassword(self.password))



##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
OpenDirectory constants.
"""

from twisted.python.constants import (
    Names, NamedConstant, Values, ValueConstant
)

from ..idirectory import (
    FieldName as BaseFieldName, RecordType as BaseRecordType
)
from ..expression import MatchType


#
# idirectory constants
#

class FieldName(Names):
    searchPath = NamedConstant()
    searchPath.description = u"search path"
    searchPath.multiValue = False

    metaNodeLocation = NamedConstant()
    metaNodeLocation.description = u"source OD node"
    metaNodeLocation.multiValue = False

    metaRecordName = NamedConstant()
    metaRecordName.description = u"meta record name"
    metaRecordName.multiValue = False

    memberUIDs = NamedConstant()
    memberUIDs.description = u"member UIDs (excluding groups)"
    memberUIDs.multiValue = True

    nestedGroupsUIDs = NamedConstant()
    nestedGroupsUIDs.description = u"nested group member UIDs"
    nestedGroupsUIDs.multiValue = True



#
# OD constants
#

class ODSearchPath(Values):
    local = ValueConstant(u"/Local/Default")
    search = ValueConstant(u"/Search")



class ODRecordType(Values):
    user = ValueConstant(u"dsRecTypeStandard:Users")
    user.recordType = BaseRecordType.user

    group = ValueConstant(u"dsRecTypeStandard:Groups")
    group.recordType = BaseRecordType.group

    # location = ValueConstant(u"dsRecTypeStandard:Locations")
    place = ValueConstant(u"dsRecTypeStandard:Places")

    resource = ValueConstant(u"dsRecTypeStandard:Resources")

    # accessControl = ValueConstant(u"dsRecTypeStandard:AccessControls")
    # afpServer = ValueConstant(u"dsRecTypeStandard:AFPServer")
    # userAlias = ValueConstant(u"dsRecTypeStandard:AFPUserAliases")
    # alias = ValueConstant(u"dsRecTypeStandard:Aliases")
    # augment = ValueConstant(u"dsRecTypeStandard:Augments")
    # autoMount = ValueConstant(u"dsRecTypeStandard:Automount")
    # autoMountMap = ValueConstant(u"dsRecTypeStandard:AutomountMap")
    # autoServerSetup = ValueConstant(u"dsRecTypeStandard:AutoServerSetup")
    # bootp = ValueConstant(u"dsRecTypeStandard:Bootp")
    # certificateAuthority = ValueConstant(u"dsRecTypeStandard:CertificateAuthorities")
    # computerList = ValueConstant(u"dsRecTypeStandard:ComputerLists")
    # computerGroup = ValueConstant(u"dsRecTypeStandard:ComputerGroups")
    # computer = ValueConstant(u"dsRecTypeStandard:Computers")
    # config = ValueConstant(u"dsRecTypeStandard:Config")
    # ethernet = ValueConstant(u"dsRecTypeStandard:Ethernets")
    # fileMakerServer = ValueConstant(u"dsRecTypeStandard:FileMakerServers")
    # ftpServer = ValueConstant(u"dsRecTypeStandard:FTPServer")
    # groupAlias = ValueConstant(u"dsRecTypeStandard:GroupAliases")
    # hostServices = ValueConstant(u"dsRecTypeStandard:HostServices")
    # host = ValueConstant(u"dsRecTypeStandard:Hosts")
    # ldapServer = ValueConstant(u"dsRecTypeStandard:LDAPServer")
    # machine = ValueConstant(u"dsRecTypeStandard:Machines")
    # map = ValueConstant(u"dsRecTypeStandard:Maps")
    # meta = ValueConstant(u"dsRecTypeStandard:AppleMetaRecord")
    # mount = ValueConstant(u"dsRecTypeStandard:Mounts")
    # neighborhood = ValueConstant(u"dsRecTypeStandard:Neighborhoods")
    # nfs = ValueConstant(u"dsRecTypeStandard:NFS")
    # netDomain = ValueConstant(u"dsRecTypeStandard:NetDomains")
    # netGroup = ValueConstant(u"dsRecTypeStandard:NetGroups")
    # network = ValueConstant(u"dsRecTypeStandard:Networks")
    # passwordServer = ValueConstant(u"dsRecTypeStandard:PasswordServer")
    # person = ValueConstant(u"dsRecTypeStandard:People")
    # presetComputer = ValueConstant(u"dsRecTypeStandard:PresetComputers")
    # presetComputerGroup = ValueConstant(u"dsRecTypeStandard:PresetComputerGroups")
    # presetComputerList = ValueConstant(u"dsRecTypeStandard:PresetComputerLists")
    # presetGroup = ValueConstant(u"dsRecTypeStandard:PresetGroups")
    # presetUser = ValueConstant(u"dsRecTypeStandard:PresetUsers")
    # printService = ValueConstant(u"dsRecTypeStandard:PrintService")
    # printServerUser = ValueConstant(u"dsRecTypeStandard:PrintServiceUser")
    # printer = ValueConstant(u"dsRecTypeStandard:Printers")
    # protocol = ValueConstant(u"dsRecTypeStandard:Protocols")
    # qtsServer = ValueConstant(u"dsRecTypeStandard:QTSServer")
    # rpc = ValueConstant(u"dsRecTypeStandard:RPC")
    # smbServer = ValueConstant(u"dsRecTypeStandard:SMBServer")
    # server = ValueConstant(u"dsRecTypeStandard:Server")
    # service = ValueConstant(u"dsRecTypeStandard:Services")
    # sharePoint = ValueConstant(u"dsRecTypeStandard:SharePoints")
    # userAlias = ValueConstant(u"dsRecTypeStandard:UserAliases")
    # webServer = ValueConstant(u"dsRecTypeStandard:WebServer")


    @classmethod
    def fromRecordType(cls, recordType):
        if not hasattr(cls, "_recordTypeByRecordType"):
            cls._recordTypeByRecordType = dict((
                (recordType.recordType, recordType)
                for recordType in cls.iterconstants()
                if hasattr(recordType, "recordType")
            ))

        return cls._recordTypeByRecordType.get(recordType, None)



class ODAttribute(Values):
    searchPath = ValueConstant(u"dsAttrTypeStandard:SearchPath")
    searchPath.fieldName = FieldName.searchPath

    recordType = ValueConstant(u"dsAttrTypeStandard:RecordType")
    recordType.fieldName = BaseFieldName.recordType

    guid = ValueConstant(u"dsAttrTypeStandard:GeneratedUID")
    guid.fieldName = BaseFieldName.guid

    shortName = ValueConstant(u"dsAttrTypeStandard:RecordName")
    shortName.fieldName = BaseFieldName.shortNames

    fullName = ValueConstant(u"dsAttrTypeStandard:RealName")
    fullName.fieldName = BaseFieldName.fullNames

    emailAddress = ValueConstant(u"dsAttrTypeStandard:EMailAddress")
    emailAddress.fieldName = BaseFieldName.emailAddresses

    metaNodeLocation = ValueConstant(u"dsAttrTypeStandard:AppleMetaNodeLocation")
    metaNodeLocation.fieldName = FieldName.metaNodeLocation

    metaRecordName = ValueConstant(u"dsAttrTypeStandard:AppleMetaRecordName")
    metaRecordName.fieldName = FieldName.metaRecordName

    groupMembers = ValueConstant(u"dsAttrTypeStandard:GroupMembers")
    groupMembers.fieldName = FieldName.memberUIDs

    nestedGroups = ValueConstant(u"dsAttrTypeStandard:NestedGroups")
    nestedGroups.fieldName = FieldName.nestedGroupsUIDs

    # For determining whether it's a system record
    uniqueId = ValueConstant(u"dsAttrTypeStandard:UniqueID")
    primaryGroupId = ValueConstant(u"dsAttrTypeStandard:PrimaryGroupID")
    isHidden = ValueConstant(u"dsAttrTypeNative:IsHidden")

    streetAddress = ValueConstant(u"dsAttrTypeStandard:Street")

    # # Single value fields
    # AdminLimits = ValueConstant(u"dsAttrTypeStandard:AdminLimits")
    # AppleAliasData = ValueConstant(u"dsAttrTypeStandard:AppleAliasData")
    # AlternateDatastoreLocation = ValueConstant(u"dsAttrTypeStandard:AlternateDatastoreLocation")
    # AuthenticationHint = ValueConstant(u"dsAttrTypeStandard:AuthenticationHint")
    # AttributeTypes = ValueConstant(u"dsAttrTypeStandard:AttributeTypes")
    # AuthorityRevocationList = ValueConstant(u"dsAttrTypeStandard:AuthorityRevocationList")
    # Birthday = ValueConstant(u"dsAttrTypeStandard:Birthday")
    # BootFile = ValueConstant(u"dsAttrTypeStandard:BootFile")
    # CACertificate = ValueConstant(u"dsAttrTypeStandard:CACertificate")
    # Capabilities = ValueConstant(u"dsAttrTypeStandard:Capabilities")
    # Capacity = ValueConstant(u"dsAttrTypeStandard:Capacity")
    # Category = ValueConstant(u"dsAttrTypeStandard:Category")
    # CertificateRevocationList = ValueConstant(u"dsAttrTypeStandard:CertificateRevocationList")
    # Change = ValueConstant(u"dsAttrTypeStandard:Change")
    # Comment = ValueConstant(u"dsAttrTypeStandard:Comment")
    # ContactGUID = ValueConstant(u"dsAttrTypeStandard:ContactGUID")
    # ContactPerson = ValueConstant(u"dsAttrTypeStandard:ContactPerson")
    # CreationTimestamp = ValueConstant(u"dsAttrTypeStandard:CreationTimestamp")
    # CrossCertificatePair = ValueConstant(u"dsAttrTypeStandard:CrossCertificatePair")
    # DataStamp = ValueConstant(u"dsAttrTypeStandard:DataStamp")
    # DNSDomain = ValueConstant(u"dsAttrTypeStandard:DNSDomain")
    # DNSNameServer = ValueConstant(u"dsAttrTypeStandard:DNSNameServer")
    # ENetAddress = ValueConstant(u"dsAttrTypeStandard:ENetAddress")
    # Expire = ValueConstant(u"dsAttrTypeStandard:Expire")
    # FirstName = ValueConstant(u"dsAttrTypeStandard:FirstName")
    # HomeDirectoryQuota = ValueConstant(u"dsAttrTypeStandard:HomeDirectoryQuota")
    # HomeDirectorySoftQuota = ValueConstant(u"dsAttrTypeStandard:HomeDirectorySoftQuota")
    # HomeLocOwner = ValueConstant(u"dsAttrTypeStandard:HomeLocOwner")
    # InetAlias = ValueConstant(u"dsAttrTypeStandard:InetAlias")
    # KDCConfigData = ValueConstant(u"dsAttrTypeStandard:KDCConfigData")
    # LastName = ValueConstant(u"dsAttrTypeStandard:LastName")
    # LDAPSearchBaseSuffix = ValueConstant(u"dsAttrTypeStandard:LDAPSearchBaseSuffix")
    # Location = ValueConstant(u"dsAttrTypeStandard:Location")
    # MapGUID = ValueConstant(u"dsAttrTypeStandard:MapGUID")
    # MCXFlags = ValueConstant(u"dsAttrTypeStandard:MCXFlags")
    # MCXSettings = ValueConstant(u"dsAttrTypeStandard:MCXSettings")
    # MailAttribute = ValueConstant(u"dsAttrTypeStandard:MailAttribute")
    # MetaAutomountMap = ValueConstant(u"dsAttrTypeStandard:MetaAutomountMap")
    # MiddleName = ValueConstant(u"dsAttrTypeStandard:MiddleName")
    # ModificationTimestamp = ValueConstant(u"dsAttrTypeStandard:ModificationTimestamp")
    # NeighborhoodAlias = ValueConstant(u"dsAttrTypeStandard:NeighborhoodAlias")
    # NeighborhoodType = ValueConstant(u"dsAttrTypeStandard:NeighborhoodType")
    # NetworkView = ValueConstant(u"dsAttrTypeStandard:NetworkView")
    # NFSHomeDirectory = ValueConstant(u"dsAttrTypeStandard:NFSHomeDirectory")
    # Note = ValueConstant(u"dsAttrTypeStandard:Note")
    # Owner = ValueConstant(u"dsAttrTypeStandard:Owner")
    # OwnerGUID = ValueConstant(u"dsAttrTypeStandard:OwnerGUID")
    # Password = ValueConstant(u"dsAttrTypeStandard:Password")
    # PasswordPlus = ValueConstant(u"dsAttrTypeStandard:PasswordPlus")
    # PasswordPolicyOptions = ValueConstant(u"dsAttrTypeStandard:PasswordPolicyOptions")
    # PasswordServerList = ValueConstant(u"dsAttrTypeStandard:PasswordServerList")
    # PasswordServerLocation = ValueConstant(u"dsAttrTypeStandard:PasswordServerLocation")
    # Picture = ValueConstant(u"dsAttrTypeStandard:Picture")
    # Port = ValueConstant(u"dsAttrTypeStandard:Port")
    # PresetUserIsAdmin = ValueConstant(u"dsAttrTypeStandard:PresetUserIsAdmin")
    # PrimaryComputerGUID = ValueConstant(u"dsAttrTypeStandard:PrimaryComputerGUID")
    # PrimaryComputerList = ValueConstant(u"dsAttrTypeStandard:PrimaryComputerList")
    # Printer1284DeviceID = ValueConstant(u"dsAttrTypeStandard:Printer1284DeviceID")
    # PrinterLPRHost = ValueConstant(u"dsAttrTypeStandard:PrinterLPRHost")
    # PrinterLPRQueue = ValueConstant(u"dsAttrTypeStandard:PrinterLPRQueue")
    # PrinterMakeAndModel = ValueConstant(u"dsAttrTypeStandard:PrinterMakeAndModel")
    # PrinterType = ValueConstant(u"dsAttrTypeStandard:PrinterType")
    # PrinterURI = ValueConstant(u"dsAttrTypeStandard:PrinterURI")
    # PrinterXRISupported = ValueConstant(u"dsAttrTypeStandard:PrinterXRISupported")
    # PrintServiceInfoText = ValueConstant(u"dsAttrTypeStandard:PrintServiceInfoText")
    # PrintServiceInfoXML = ValueConstant(u"dsAttrTypeStandard:PrintServiceInfoXML")
    # PrintServiceUserData = ValueConstant(u"dsAttrTypeStandard:PrintServiceUserData")
    # RealUserID = ValueConstant(u"dsAttrTypeStandard:RealUserID")
    # RelativeDNPrefix = ValueConstant(u"dsAttrTypeStandard:RelativeDNPrefix")
    # SMBAccountFlags = ValueConstant(u"dsAttrTypeStandard:SMBAccountFlags")
    # SMBGroupRID = ValueConstant(u"dsAttrTypeStandard:SMBGroupRID")
    # SMBHome = ValueConstant(u"dsAttrTypeStandard:SMBHome")
    # SMBHomeDrive = ValueConstant(u"dsAttrTypeStandard:SMBHomeDrive")
    # SMBKickoffTime = ValueConstant(u"dsAttrTypeStandard:SMBKickoffTime")
    # SMBLogoffTime = ValueConstant(u"dsAttrTypeStandard:SMBLogoffTime")
    # SMBLogonTime = ValueConstant(u"dsAttrTypeStandard:SMBLogonTime")
    # SMBPrimaryGroupSID = ValueConstant(u"dsAttrTypeStandard:SMBPrimaryGroupSID")
    # SMBPasswordLastSet = ValueConstant(u"dsAttrTypeStandard:SMBPasswordLastSet")
    # SMBProfilePath = ValueConstant(u"dsAttrTypeStandard:SMBProfilePath")
    # SMBRID = ValueConstant(u"dsAttrTypeStandard:SMBRID")
    # SMBScriptPath = ValueConstant(u"dsAttrTypeStandard:SMBScriptPath")
    # SMBSID = ValueConstant(u"dsAttrTypeStandard:SMBSID")
    # SMBUserWorkstations = ValueConstant(u"dsAttrTypeStandard:SMBUserWorkstations")
    # ServiceType = ValueConstant(u"dsAttrTypeStandard:ServiceType")
    # SetupAssistantAdvertising = ValueConstant(u"dsAttrTypeStandard:SetupAssistantAdvertising")
    # SetupAssistantAutoRegister = ValueConstant(u"dsAttrTypeStandard:SetupAssistantAutoRegister")
    # SetupAssistantLocation = ValueConstant(u"dsAttrTypeStandard:SetupAssistantLocation")
    # Occupation = ValueConstant(u"dsAttrTypeStandard:Occupation")
    # TimeToLive = ValueConstant(u"dsAttrTypeStandard:TimeToLive")
    # UserCertificate = ValueConstant(u"dsAttrTypeStandard:UserCertificate")
    # UserPKCS12Data = ValueConstant(u"dsAttrTypeStandard:UserPKCS12Data")
    # UserShell = ValueConstant(u"dsAttrTypeStandard:UserShell")
    # UserSMIMECertificate = ValueConstant(u"dsAttrTypeStandard:UserSMIMECertificate")
    # VFSDumpFreq = ValueConstant(u"dsAttrTypeStandard:VFSDumpFreq")
    # VFSLinkDir = ValueConstant(u"dsAttrTypeStandard:VFSLinkDir")
    # VFSPassNo = ValueConstant(u"dsAttrTypeStandard:VFSPassNo")
    # VFSType = ValueConstant(u"dsAttrTypeStandard:VFSType")
    # WeblogURI = ValueConstant(u"dsAttrTypeStandard:WeblogURI")
    # XMLPlist = ValueConstant(u"dsAttrTypeStandard:XMLPlist")
    # ProtocolNumber = ValueConstant(u"dsAttrTypeStandard:ProtocolNumber")
    # RPCNumber = ValueConstant(u"dsAttrTypeStandard:RPCNumber")
    # NetworkNumber = ValueConstant(u"dsAttrTypeStandard:NetworkNumber")

    # # Multi-value fields
    # AccessControlEntry = ValueConstant(u"dsAttrTypeStandard:AccessControlEntry")
    # AddressLine1 = ValueConstant(u"dsAttrTypeStandard:AddressLine1")
    # AddressLine2 = ValueConstant(u"dsAttrTypeStandard:AddressLine2")
    # AddressLine3 = ValueConstant(u"dsAttrTypeStandard:AddressLine3")
    # AltSecurityIdentities = ValueConstant(u"dsAttrTypeStandard:AltSecurityIdentities")
    # AreaCode = ValueConstant(u"dsAttrTypeStandard:AreaCode")
    # AuthenticationAuthority = ValueConstant(u"dsAttrTypeStandard:AuthenticationAuthority")
    # AutomountInformation = ValueConstant(u"dsAttrTypeStandard:AutomountInformation")
    # BootParams = ValueConstant(u"dsAttrTypeStandard:BootParams")
    # Building = ValueConstant(u"dsAttrTypeStandard:Building")
    # CalendarPrincipalURI = ValueConstant(u"dsAttrTypeStandard:CalendarPrincipalURI")
    # City = ValueConstant(u"dsAttrTypeStandard:City")
    # Company = ValueConstant(u"dsAttrTypeStandard:Company")
    # ComputerAlias = ValueConstant(u"dsAttrTypeStandard:ComputerAlias")
    # Computers = ValueConstant(u"dsAttrTypeStandard:Computers")
    # Country = ValueConstant(u"dsAttrTypeStandard:Country")
    # Department = ValueConstant(u"dsAttrTypeStandard:Department")
    # DNSName = ValueConstant(u"dsAttrTypeStandard:DNSName")
    # EMailContacts = ValueConstant(u"dsAttrTypeStandard:EMailContacts")
    # FAXNumber = ValueConstant(u"dsAttrTypeStandard:FAXNumber")
    # Group = ValueConstant(u"dsAttrTypeStandard:Group")
    # GroupMembers = ValueConstant(u"dsAttrTypeStandard:GroupMembers")
    # GroupMembership = ValueConstant(u"dsAttrTypeStandard:GroupMembership")
    # GroupServices = ValueConstant(u"dsAttrTypeStandard:GroupServices")
    # HomePhoneNumber = ValueConstant(u"dsAttrTypeStandard:HomePhoneNumber")
    # HTML = ValueConstant(u"dsAttrTypeStandard:HTML")
    # HomeDirectory = ValueConstant(u"dsAttrTypeStandard:HomeDirectory")
    # IMHandle = ValueConstant(u"dsAttrTypeStandard:IMHandle")
    # IPAddress = ValueConstant(u"dsAttrTypeStandard:IPAddress")
    # IPAddressAndENetAddress = ValueConstant(u"dsAttrTypeStandard:IPAddressAndENetAddress")
    # IPv6Address = ValueConstant(u"dsAttrTypeStandard:IPv6Address")
    # JPEGPhoto = ValueConstant(u"dsAttrTypeStandard:JPEGPhoto")
    # JobTitle = ValueConstant(u"dsAttrTypeStandard:JobTitle")
    # KDCAuthKey = ValueConstant(u"dsAttrTypeStandard:KDCAuthKey")
    # Keywords = ValueConstant(u"dsAttrTypeStandard:Keywords")
    # LDAPReadReplicas = ValueConstant(u"dsAttrTypeStandard:LDAPReadReplicas")
    # LDAPWriteReplicas = ValueConstant(u"dsAttrTypeStandard:LDAPWriteReplicas")
    # MachineServes = ValueConstant(u"dsAttrTypeStandard:MachineServes")
    # MapCoordinates = ValueConstant(u"dsAttrTypeStandard:MapCoordinates")
    # MapURI = ValueConstant(u"dsAttrTypeStandard:MapURI")
    # MCXSettings = ValueConstant(u"dsAttrTypeStandard:MCXSettings")
    # MIME = ValueConstant(u"dsAttrTypeStandard:MIME")
    # Member = ValueConstant(u"dsAttrTypeStandard:Member")
    # MobileNumber = ValueConstant(u"dsAttrTypeStandard:MobileNumber")
    # NBPEntry = ValueConstant(u"dsAttrTypeStandard:NBPEntry")
    # NestedGroups = ValueConstant(u"dsAttrTypeStandard:NestedGroups")
    # NetGroups = ValueConstant(u"dsAttrTypeStandard:NetGroups")
    # NickName = ValueConstant(u"dsAttrTypeStandard:NickName")
    # NodePathXMLPlist = ValueConstant(u"dsAttrTypeStandard:NodePathXMLPlist")
    # OrganizationInfo = ValueConstant(u"dsAttrTypeStandard:OrganizationInfo")
    # OrganizationName = ValueConstant(u"dsAttrTypeStandard:OrganizationName")
    # PagerNumber = ValueConstant(u"dsAttrTypeStandard:PagerNumber")
    # PhoneContacts = ValueConstant(u"dsAttrTypeStandard:PhoneContacts")
    # PhoneNumber = ValueConstant(u"dsAttrTypeStandard:PhoneNumber")
    # PGPPublicKey = ValueConstant(u"dsAttrTypeStandard:PGPPublicKey")
    # PostalAddress = ValueConstant(u"dsAttrTypeStandard:PostalAddress")
    # PostalAddressContacts = ValueConstant(u"dsAttrTypeStandard:PostalAddressContacts")
    # PostalCode = ValueConstant(u"dsAttrTypeStandard:PostalCode")
    # NamePrefix = ValueConstant(u"dsAttrTypeStandard:NamePrefix")
    # Protocols = ValueConstant(u"dsAttrTypeStandard:Protocols")
    # Relationships = ValueConstant(u"dsAttrTypeStandard:Relationships")
    # ResourceInfo = ValueConstant(u"dsAttrTypeStandard:ResourceInfo")
    # ResourceType = ValueConstant(u"dsAttrTypeStandard:ResourceType")
    # ServicesLocator = ValueConstant(u"dsAttrTypeStandard:ServicesLocator")
    # State = ValueConstant(u"dsAttrTypeStandard:State")
    # Street = ValueConstant(u"dsAttrTypeStandard:Street")
    # NameSuffix = ValueConstant(u"dsAttrTypeStandard:NameSuffix")
    # URL = ValueConstant(u"dsAttrTypeStandard:URL")
    # URLForNSL = ValueConstant(u"dsAttrTypeStandard:URLForNSL")
    # VFSOpts = ValueConstant(u"dsAttrTypeStandard:VFSOpts")

    # # Other fields
    # AdminStatus = ValueConstant(u"dsAttrTypeStandard:AdminStatus")
    # Alias = ValueConstant(u"dsAttrTypeStandard:Alias")
    # AuthCredential = ValueConstant(u"dsAttrTypeStandard:AuthCredential")
    # CopyTimestamp = ValueConstant(u"dsAttrTypeStandard:CopyTimestamp")
    # DateRecordCreated = ValueConstant(u"dsAttrTypeStandard:DateRecordCreated")
    # KerberosRealm = ValueConstant(u"dsAttrTypeStandard:KerberosRealm")
    # NTDomainComputerAccount = ValueConstant(u"dsAttrTypeStandard:NTDomainComputerAccount")
    # OriginalHomeDirectory = ValueConstant(u"dsAttrTypeStandard:OriginalHomeDirectory")
    # OriginalNFSHomeDirectory = ValueConstant(u"dsAttrTypeStandard:OriginalNFSHomeDirectory")
    # OriginalNodeName = ValueConstant(u"dsAttrTypeStandard:OriginalNodeName")
    # PrimaryNTDomain = ValueConstant(u"dsAttrTypeStandard:PrimaryNTDomain")
    # PwdAgingPolicy = ValueConstant(u"dsAttrTypeStandard:PwdAgingPolicy")
    # RARA = ValueConstant(u"dsAttrTypeStandard:RARA")
    # ReadOnlyNode = ValueConstant(u"dsAttrTypeStandard:ReadOnlyNode")
    # RecordImage = ValueConstant(u"dsAttrTypeStandard:RecordImage")
    # SMBGroupRID = ValueConstant(u"dsAttrTypeStandard:SMBGroupRID")
    # TimePackage = ValueConstant(u"dsAttrTypeStandard:TimePackage")
    # TotalSize = ValueConstant(u"dsAttrTypeStandard:TotalSize")
    # AllNames = ValueConstant(u"dsAttrTypeStandard:AllNames")
    # AuthMethod = ValueConstant(u"dsAttrTypeStandard:AuthMethod")
    # NodePath = ValueConstant(u"dsAttrTypeStandard:NodePath")
    # PlugInInfo = ValueConstant(u"dsAttrTypeStandard:PlugInInfo")
    # RecordAlias = ValueConstant(u"dsAttrTypeStandard:RecordAlias")
    # Scheama = ValueConstant(u"dsAttrTypeStandard:Scheama")
    # SetPasswdMethod = ValueConstant(u"dsAttrTypeStandard:SetPasswdMethod")
    # SubNodes = ValueConstant(u"dsAttrTypeStandard:SubNodes")
    # AppleMetaAliasSource = ValueConstant(u"dsAttrTypeStandard:AppleMetaAliasSource")
    # AppleMetaAliasTarget = ValueConstant(u"dsAttrTypeStandard:AppleMetaAliasTarget")
    # NetGroupTriplet = ValueConstant(u"dsAttrTypeStandard:NetGroupTriplet")
    # SearchPolicy = ValueConstant(u"dsAttrTypeStandard:SearchPolicy")
    # NSPSearchPath = ValueConstant(u"dsAttrTypeStandard:NSPSearchPath")
    # LSPSearchPath = ValueConstant(u"dsAttrTypeStandard:LSPSearchPath")
    # CSPSearchPath = ValueConstant(u"dsAttrTypeStandard:CSPSearchPath")


    @classmethod
    def fromFieldName(cls, fieldName):
        if not hasattr(cls, "_attributesByFieldName"):
            cls._attributesByFieldName = dict((
                (attribute.fieldName, attribute)
                for attribute in cls.iterconstants()
                if hasattr(attribute, "fieldName")
            ))

        return cls._attributesByFieldName.get(fieldName, None)



class ODMatchType(Values):
    any = ValueConstant(0x0001)
    any.queryString = u"({notOp}{attribute}=*)"

    equals = ValueConstant(0x2001)
    equals.matchType = MatchType.equals
    equals.queryString = u"({notOp}{attribute}={value})"

    startsWith = ValueConstant(0x2002)
    startsWith.matchType = MatchType.startsWith
    startsWith.queryString = u"({notOp}{attribute}={value}*)"

    endsWith = ValueConstant(0x2003)
    endsWith.matchType = MatchType.endsWith
    endsWith.queryString = u"({notOp}{attribute}=*{value})"

    contains = ValueConstant(0x2004)
    contains.matchType = MatchType.contains
    contains.queryString = u"({notOp}{attribute}=*{value}*)"

    lessThan = ValueConstant(0x2005)
    lessThan.matchType = MatchType.lessThan
    lessThan.queryString = u"({notOp}{attribute}<{value})"

    greaterThan = ValueConstant(0x2006)
    greaterThan.matchType = MatchType.greaterThan
    greaterThan.queryString = u"({notOp}{attribute}>{value})"

    lessThanOrEqualTo = ValueConstant(0x2007)
    lessThanOrEqualTo.matchType = MatchType.lessThanOrEqualTo
    lessThanOrEqualTo.queryString = u"({notOp}{attribute}<={value})"

    greaterThanOrEqualTo = ValueConstant(0x2008)
    greaterThanOrEqualTo.matchType = MatchType.greaterThanOrEqualTo
    greaterThanOrEqualTo.queryString = u"({notOp}{attribute}>={value})"

    compound = ValueConstant(0x210B)


    @classmethod
    def fromMatchType(cls, matchType):
        if not hasattr(cls, "_matchTypeByMatchType"):
            cls._matchTypeByMatchType = dict((
                (matchType.matchType, matchType)
                for matchType in cls.iterconstants()
                if hasattr(matchType, "matchType")
            ))

        return cls._matchTypeByMatchType.get(matchType, None)



class ODAuthMethod(Values):
    digestMD5 = ValueConstant(u"dsAuthMethodStandard:dsAuthNodeDIGEST-MD5")



##
# Copyright (c) 2010-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Tests for L{twext.enterprise.adbapi2}.
"""

import gc

from zope.interface.verify import verifyObject

from twisted.python.failure import Failure

from twisted.trial.unittest import TestCase

from twisted.internet.defer import Deferred, fail, succeed, inlineCallbacks

from twisted.test.proto_helpers import StringTransport

from twext.enterprise.ienterprise import ConnectionError
from twext.enterprise.ienterprise import AlreadyFinishedError
from twext.enterprise.adbapi2 import ConnectionPoolClient
from twext.enterprise.adbapi2 import ConnectionPoolConnection
from twext.enterprise.ienterprise import IAsyncTransaction
from twext.enterprise.ienterprise import ICommandBlock
from twext.enterprise.adbapi2 import FailsafeException
from twext.enterprise.adbapi2 import ConnectionPool
from twext.enterprise.fixtures import ConnectionPoolHelper
from twext.enterprise.fixtures import resultOf
from twext.enterprise.fixtures import ClockWithThreads
from twext.enterprise.fixtures import FakeConnectionError
from twext.enterprise.fixtures import RollbackFail
from twext.enterprise.fixtures import CommitFail
from twext.enterprise.adbapi2 import Commit
from twext.enterprise.adbapi2 import _HookableOperation



class TrashCollector(object):
    """
    Test helper for monitoring gc.garbage.
    """
    def __init__(self, testCase):
        self.testCase = testCase
        testCase.addCleanup(self.checkTrash)
        self.start()


    def start(self):
        gc.collect()
        self.garbageStart = len(gc.garbage)


    def checkTrash(self):
        """
        Ensure that the test has added no additional garbage.
        """
        gc.collect()
        newGarbage = gc.garbage[self.garbageStart:]
        if newGarbage:
            # Don't clean up twice.
            self.start()
            self.testCase.fail("New garbage: " + repr(newGarbage))



class AssertResultHelper(object):
    """
    Mixin for asserting about synchronous Deferred results.
    """

    def assertResultList(self, resultList, expected):
        """
        Assert that a list created with L{resultOf} contais the expected
        result.

        @param resultList: The return value of L{resultOf}.
        @type resultList: L{list}

        @param expected: The expected value that should be present in the list;
            a L{Failure} if an exception is expected to be raised.
        """
        if not resultList:
            self.fail("No result; Deferred didn't fire yet.")
        else:
            if isinstance(resultList[0], Failure):
                if isinstance(expected, Failure):
                    resultList[0].trap(expected.type)
                else:
                    resultList[0].raiseException()
            else:
                self.assertEqual(resultList, [expected])



class ConnectionPoolBootTests(TestCase):
    """
    Tests for the start-up phase of L{ConnectionPool}.
    """

    def test_threadCount(self):
        """
        The reactor associated with a L{ConnectionPool} will have its maximum
        thread count adjusted when L{ConnectionPool.startService} is called, to
        accomodate for L{ConnectionPool.maxConnections} additional threads.

        Stopping the service should restore it to its original value, so that a
        repeatedly re-started L{ConnectionPool} will not cause the thread
        ceiling to grow without bound.
        """
        defaultMax = 27
        connsMax = 45
        combinedMax = defaultMax + connsMax
        pool = ConnectionPool(None, maxConnections=connsMax)
        pool.reactor = ClockWithThreads()
        threadpool = pool.reactor.getThreadPool()
        pool.reactor.suggestThreadPoolSize(defaultMax)
        self.assertEquals(threadpool.max, defaultMax)
        pool.startService()
        self.assertEquals(threadpool.max, combinedMax)
        justChecking = []
        pool.stopService().addCallback(justChecking.append)
        # No SQL run, so no threads started, so this deferred should fire
        # immediately.  If not, we're in big trouble, so sanity check.
        self.assertEquals(justChecking, [None])
        self.assertEquals(threadpool.max, defaultMax)


    def test_isRunning(self):
        """
        L{ConnectionPool.startService} should set its C{running} attribute to
        true.
        """
        pool = ConnectionPool(None)
        pool.reactor = ClockWithThreads()
        self.assertEquals(pool.running, False)
        pool.startService()
        self.assertEquals(pool.running, True)



class ConnectionPoolNameTests(TestCase):
    """
    Tests for L{ConnectionPool}'s C{name} attribute.
    """
    def test_default(self):
        """
        If no value is given for the C{name} parameter to L{ConnectionPool}'s
        initializer then L{ConnectionPool.name} is C{None}.
        """
        pool = ConnectionPool(None)
        self.assertIs(None, pool.name)


    def test_specified(self):
        """
        If a value is given for the C{name} parameter to L{ConnectionPool}'s
        initializer then it is used as the value for L{ConnectionPool.name}.
        """
        name = "some test pool"
        pool = ConnectionPool(None, name=name)
        self.assertEqual(name, pool.name)



class ConnectionPoolTests(ConnectionPoolHelper, TestCase, AssertResultHelper):
    """
    Tests for L{ConnectionPool}.
    """

    def test_tooManyConnections(self):
        """
        When the number of outstanding busy transactions exceeds the number of
        slots specified by L{ConnectionPool.maxConnections},
        L{ConnectionPool.connection} will return a pooled transaction that is
        not backed by any real database connection; this object will queue its
        SQL statements until an existing connection becomes available.
        """
        a = self.createTransaction()

        alphaResult = self.resultOf(a.execSQL("alpha"))
        [[_ignore_counter, _ignore_echo]] = alphaResult[0]

        b = self.createTransaction()
        # "b" should have opened a connection.
        self.assertEquals(len(self.factory.connections), 2)
        betaResult = self.resultOf(b.execSQL("beta"))
        [[bcounter, _ignore_becho]] = betaResult[0]

        # both "a" and "b" are holding open a connection now; let's try to open
        # a third one.  (The ordering will be deterministic even if this fails,
        # because those threads are already busy.)
        c = self.createTransaction()
        gammaResult = self.resultOf(c.execSQL("gamma"))

        # Did "c" open a connection?  Let's hope not...
        self.assertEquals(len(self.factory.connections), 2)
        # SQL shouldn't be executed too soon...
        self.assertEquals(gammaResult, [])

        commitResult = self.resultOf(b.commit())

        # Now that "b" has committed, "c" should be able to complete.
        [[ccounter, _ignore_cecho]] = gammaResult[0]

        # The connection for "a" ought to still be busy, so let's make sure
        # we're using the one for "c".
        self.assertEquals(ccounter, bcounter)

        # Sanity check: the commit should have succeeded!
        self.assertEquals(commitResult, [None])


    def test_stopService(self):
        """
        L{ConnectionPool.stopService} stops all the associated L{ThreadHolder}s
        and thereby frees up the resources it is holding.
        """
        a = self.createTransaction()
        alphaResult = self.resultOf(a.execSQL("alpha"))
        [[[_ignore_counter, _ignore_echo]]] = alphaResult
        self.assertEquals(len(self.factory.connections), 1)
        self.assertEquals(len(self.holders), 1)
        [holder] = self.holders
        self.assertEquals(holder.started, True)
        self.assertEquals(holder.stopped, False)
        self.pool.stopService()
        self.assertEquals(self.pool.running, False)
        self.assertEquals(len(self.holders), 1)
        self.assertEquals(holder.started, True)
        self.assertEquals(holder.stopped, True)
        # Closing fake connections removes them from the list.
        self.assertEquals(len(self.factory.connections), 1)
        self.assertEquals(self.factory.connections[0].closed, True)


    def test_retryAfterConnectError(self):
        """
        When the C{connectionFactory} passed to L{ConnectionPool} raises an
        exception, the L{ConnectionPool} will log the exception and delay
        execution of a new connection's SQL methods until an attempt succeeds.
        """
        self.factory.willFail()
        self.factory.willFail()
        self.factory.willConnect()
        c = self.createTransaction()

        def checkOneFailure():
            errors = self.flushLoggedErrors(FakeConnectionError)
            self.assertEquals(len(errors), 1)

        checkOneFailure()
        d = c.execSQL("alpha")
        happened = []
        d.addBoth(happened.append)
        self.assertEquals(happened, [])
        self.clock.advance(self.pool.RETRY_TIMEOUT + 0.01)
        checkOneFailure()
        self.assertEquals(happened, [])
        self.clock.advance(self.pool.RETRY_TIMEOUT + 0.01)
        self.flushHolders()
        self.assertEquals(happened, [[[1, "alpha"]]])


    def test_shutdownDuringRetry(self):
        """
        If a L{ConnectionPool} is attempting to shut down while it's in the
        process of re-trying a connection attempt that received an error, the
        connection attempt should be cancelled and the shutdown should complete
        as normal.
        """
        self.factory.defaultFail()
        self.createTransaction()
        errors = self.flushLoggedErrors(FakeConnectionError)
        self.assertEquals(len(errors), 1)
        stopd = []
        self.pool.stopService().addBoth(stopd.append)
        self.assertResultList(stopd, None)
        self.assertEquals(self.clock.calls, [])
        [holder] = self.holders
        self.assertEquals(holder.started, True)
        self.assertEquals(holder.stopped, True)


    def test_shutdownDuringAttemptSuccess(self):
        """
        If L{ConnectionPool.stopService} is called while a connection attempt
        is outstanding, the resulting L{Deferred} won't be fired until the
        connection attempt has finished; in this case, succeeded.
        """
        self.pauseHolders()
        self.createTransaction()
        stopd = []
        self.pool.stopService().addBoth(stopd.append)
        self.assertEquals(stopd, [])
        self.flushHolders()
        self.assertResultList(stopd, None)
        [holder] = self.holders
        self.assertEquals(holder.started, True)
        self.assertEquals(holder.stopped, True)


    def test_shutdownDuringAttemptFailed(self):
        """
        If L{ConnectionPool.stopService} is called while a connection attempt
        is outstanding, the resulting L{Deferred} won't be fired until the
        connection attempt has finished; in this case, failed.
        """
        self.factory.defaultFail()
        self.pauseHolders()
        self.createTransaction()
        stopd = []
        self.pool.stopService().addBoth(stopd.append)
        self.assertEquals(stopd, [])
        self.flushHolders()
        errors = self.flushLoggedErrors(FakeConnectionError)
        self.assertEquals(len(errors), 1)
        self.assertResultList(stopd, None)
        [holder] = self.holders
        self.assertEquals(holder.started, True)
        self.assertEquals(holder.stopped, True)


    def test_stopServicePseudoTxn(self):
        """
        When L{ConnectionPool.stopService} is called with a pending
        L{_ConnectingPseudoTxn} active, the DB connection being created
        is closed.
        """
        self.pauseHolders()
        self.createTransaction()
        stopResult = self.resultOf(self.pool.stopService())
        self.assertEquals(stopResult, [])
        self.flushHolders()
        [holder] = self.holders
        self.assertEquals(holder.started, True)
        self.assertEquals(holder.stopped, True)
        self.assertEquals(len(self.factory.connections), 1)
        self.assertEquals(self.factory.connections[0].closed, True)


    def test_stopServiceMidAbort(self):
        """
        When L{ConnectionPool.stopService} is called with deferreds from
        C{abort} still outstanding, it will wait for the currently-aborting
        transaction to fully abort before firing the L{Deferred} returned from
        C{stopService}.
        """
        # TODO: commit() too?
        self.pauseHolders()
        c = self.createTransaction()
        abortResult = self.resultOf(c.abort())
        # Should abort instantly, as it hasn't managed to unspool anything yet.
        # FIXME: kill all Deferreds associated with this thing, make sure that
        # any outstanding query callback chains get nuked.
        self.assertEquals(abortResult, [None])
        stopResult = self.resultOf(self.pool.stopService())
        self.assertEquals(stopResult, [])
        self.flushHolders()
        # self.assertEquals(abortResult, [None])
        self.assertResultList(stopResult, None)


    def test_stopServiceWithSpooled(self):
        """
        When L{ConnectionPool.stopService} is called when spooled transactions
        are outstanding, any pending L{Deferreds} returned by those
        transactions will be failed with L{ConnectionError}.
        """
        # Use up the free slots so we have to spool.
        hold = []
        hold.append(self.createTransaction())
        hold.append(self.createTransaction())

        c = self.createTransaction()
        se = self.resultOf(c.execSQL("alpha"))
        ce = self.resultOf(c.commit())
        self.assertEquals(se, [])
        self.assertEquals(ce, [])
        self.resultOf(self.pool.stopService())
        self.assertEquals(se[0].type, self.translateError(ConnectionError))
        self.assertEquals(ce[0].type, self.translateError(ConnectionError))


    def test_repoolSpooled(self):
        """
        Regression test for a somewhat tricky-to-explain bug: when a spooled
        transaction which has already had commit() called on it before it's
        received a real connection to start executing on, it will not leave
        behind any detritus that prevents stopService from working.
        """
        self.pauseHolders()
        c = self.createTransaction()
        c2 = self.createTransaction()
        c3 = self.createTransaction()
        c.commit()
        c2.commit()
        c3.commit()
        self.flushHolders()
        self.assertEquals(len(self.factory.connections), 2)
        stopResult = self.resultOf(self.pool.stopService())
        self.assertEquals(stopResult, [None])
        self.assertEquals(len(self.factory.connections), 2)
        self.assertEquals(self.factory.connections[0].closed, True)
        self.assertEquals(self.factory.connections[1].closed, True)


    def test_connectAfterStop(self):
        """
        Calls to connection() after stopService() result in transactions which
        immediately fail all operations.
        """
        stopResults = self.resultOf(self.pool.stopService())
        self.assertEquals(stopResults, [None])
        self.pauseHolders()
        postClose = self.createTransaction()
        queryResult = self.resultOf(postClose.execSQL("hello"))
        self.assertEquals(len(queryResult), 1)
        self.assertEquals(queryResult[0].type,
                          self.translateError(ConnectionError))


    def test_connectAfterStartedStopping(self):
        """
        Calls to connection() after stopService() has been called but before it
        has completed will result in transactions which immediately fail all
        operations.
        """
        self.pauseHolders()
        preClose = self.createTransaction()
        preCloseResult = self.resultOf(preClose.execSQL("statement"))
        stopResult = self.resultOf(self.pool.stopService())
        postClose = self.createTransaction()
        queryResult = self.resultOf(postClose.execSQL("hello"))
        self.assertEquals(stopResult, [])
        self.assertEquals(len(queryResult), 1)
        self.assertEquals(
            queryResult[0].type,
            self.translateError(ConnectionError)
        )
        self.assertEquals(len(preCloseResult), 1)
        self.assertEquals(
            preCloseResult[0].type,
            self.translateError(ConnectionError)
        )


    def test_abortFailsDuringStopService(self):
        """
        L{IAsyncTransaction.abort} might fail, most likely because the
        underlying database connection has already been disconnected.  If this
        happens, shutdown should continue.
        """
        txns = []
        txns.append(self.createTransaction())
        txns.append(self.createTransaction())
        for txn in txns:
            # Make sure rollback will actually be executed.
            results = self.resultOf(txn.execSQL("maybe change something!"))
            [[[_ignore_counter, echo]]] = results
            self.assertEquals("maybe change something!", echo)
        # Fail one (and only one) call to rollback().
        self.factory.rollbackFail = True
        stopResult = self.resultOf(self.pool.stopService())
        self.assertEquals(stopResult, [None])
        self.assertEquals(len(self.flushLoggedErrors(RollbackFail)), 1)
        self.assertEquals(self.factory.connections[0].closed, True)
        self.assertEquals(self.factory.connections[1].closed, True)


    def test_partialTxnFailsDuringStopService(self):
        """
        Using the logic in L{ConnectionPool.stopService}, make sure that an
        L{_ConnectedTxn} cannot continue to process SQL after L{_ConnectedTxn.abort}
        is called and before L{_ConnectedTxn.reset} is called.
        """
        txn = self.createTransaction()
        if hasattr(txn, "_baseTxn"):
            # Send initial statement
            txn.execSQL("maybe change something!")

            # Make it look like the service is stopping
            txn._baseTxn._connection.close()
            txn._baseTxn.terminate()

            # Try to send more SQL - must fail
            self.failUnlessRaises(RuntimeError, txn.execSQL, "maybe change something else!")


    def test_abortRecycledTransaction(self):
        """
        L{ConnectionPool.stopService} will shut down if a recycled transaction
        is still pending.
        """
        recycled = self.createTransaction()
        self.resultOf(recycled.commit())
        remember = []
        remember.append(self.createTransaction())
        self.assertEquals(self.resultOf(self.pool.stopService()), [None])


    def test_abortSpooled(self):
        """
        Aborting a still-spooled transaction (one which has no statements being
        executed) will result in all of its Deferreds immediately failing and
        none of the queued statements being executed.
        """
        active = []
        # Use up the available connections ...
        for _ignore in xrange(self.pool.maxConnections):
            active.append(self.createTransaction())

        # ... so that this one has to be spooled.
        spooled = self.createTransaction()
        result = self.resultOf(spooled.execSQL("alpha"))

        # sanity check, it would be bad if this actually executed.
        self.assertEqual(result, [])
        self.resultOf(spooled.abort())
        self.assertEqual(result[0].type, self.translateError(ConnectionError))


    def test_waitForAlreadyAbortedTransaction(self):
        """
        L{ConnectionPool.stopService} will wait for all transactions to shut
        down before exiting, including those which have already been stopped.
        """
        it = self.createTransaction()
        self.pauseHolders()
        abortResult = self.resultOf(it.abort())

        # steal it from the queue so we can do it out of order
        d, _ignore_work = self.holders[0]._q.get()

        # that should be the only work unit so don't continue if something else
        # got in there
        self.assertEquals(list(self.holders[0]._q.queue), [])
        self.assertEquals(len(self.holders), 1)
        self.flushHolders()
        stopResult = self.resultOf(self.pool.stopService())

        # Sanity check that we haven't actually stopped it yet
        self.assertEquals(abortResult, [])

        # We haven't fired it yet, so the service had better not have
        # stopped...
        self.assertEquals(stopResult, [])

        d.callback(None)
        self.flushHolders()
        self.assertEquals(abortResult, [None])
        self.assertEquals(stopResult, [None])


    def test_garbageCollectedTransactionAborts(self):
        """
        When an L{IAsyncTransaction} is garbage collected, it ought to abort
        itself.
        """
        t = self.createTransaction()
        self.resultOf(t.execSQL("echo", []))
        conns = self.factory.connections
        self.assertEquals(len(conns), 1)
        self.assertEquals(conns[0]._rollbackCount, 0)
        del t
        gc.collect()
        self.flushHolders()
        self.assertEquals(len(conns), 1)
        self.assertEquals(conns[0]._rollbackCount, 1)
        self.assertEquals(conns[0]._commitCount, 0)


    def circularReferenceTest(self, finish, hook):
        """
        Collecting a completed (committed or aborted) L{IAsyncTransaction}
        should not leak any circular references.
        """
        tc = TrashCollector(self)
        commitExecuted = []

        def carefullyManagedScope():
            t = self.createTransaction()

            def holdAReference():
                """
                This is a hook that holds a reference to "t".
                """
                commitExecuted.append(True)
                return t.execSQL("teardown", [])

            hook(t, holdAReference)
            finish(t)

        self.failIf(commitExecuted, "Commit hook executed.")
        carefullyManagedScope()
        tc.checkTrash()


    def test_noGarbageOnCommit(self):
        """
        Committing a transaction does not cause gc garbage.
        """
        self.circularReferenceTest(
            lambda txn: txn.commit(),
            lambda txn, hook: txn.preCommit(hook)
        )


    def test_noGarbageOnCommitWithAbortHook(self):
        """
        Committing a transaction does not cause gc garbage.
        """
        self.circularReferenceTest(
            lambda txn: txn.commit(),
            lambda txn, hook: txn.postAbort(hook)
        )


    def test_noGarbageOnAbort(self):
        """
        Aborting a transaction does not cause gc garbage.
        """
        self.circularReferenceTest(
            lambda txn: txn.abort(),
            lambda txn, hook: txn.preCommit(hook)
        )


    def test_noGarbageOnAbortWithPostCommitHook(self):
        """
        Aborting a transaction does not cause gc garbage.
        """
        self.circularReferenceTest(
            lambda txn: txn.abort(),
            lambda txn, hook: txn.postCommit(hook)
        )


    def test_tooManyConnectionsWhileOthersFinish(self):
        """
        L{ConnectionPool.connection} will not spawn more than the maximum
        connections if there are finishing transactions outstanding.
        """
        a = self.createTransaction()
        b = self.createTransaction()
        self.pauseHolders()
        a.abort()
        b.abort()

        # Remove the holders for the existing connections, so that the "extra"
        # connection() call wins the race and gets executed first.
        self.holders[:] = []
        self.createTransaction()
        self.flushHolders()
        self.assertEquals(len(self.factory.connections), 2)


    def setParamstyle(self, paramstyle):
        """
        Change the paramstyle of the transaction under test.
        """
        self.pool.paramstyle = paramstyle


    def test_propagateParamstyle(self):
        """
        Each different type of L{ISQLExecutor} relays the C{paramstyle}
        attribute from the L{ConnectionPool}.
        """
        TEST_PARAMSTYLE = "justtesting"
        self.setParamstyle(TEST_PARAMSTYLE)
        normaltxn = self.createTransaction()
        self.assertEquals(normaltxn.paramstyle, TEST_PARAMSTYLE)
        self.assertEquals(normaltxn.commandBlock().paramstyle, TEST_PARAMSTYLE)
        self.pauseHolders()
        extra = []
        extra.append(self.createTransaction())
        waitingtxn = self.createTransaction()
        self.assertEquals(waitingtxn.paramstyle, TEST_PARAMSTYLE)
        self.flushHolders()
        self.pool.stopService()
        notxn = self.createTransaction()
        self.assertEquals(notxn.paramstyle, TEST_PARAMSTYLE)


    def setDialect(self, dialect):
        """
        Change the dialect of the transaction under test.
        """
        self.pool.dialect = dialect


    def test_propagateDialect(self):
        """
        Each different type of L{ISQLExecutor} relays the C{dialect}
        attribute from the L{ConnectionPool}.
        """
        TEST_DIALECT = "otherdialect"
        self.setDialect(TEST_DIALECT)
        normaltxn = self.createTransaction()
        self.assertEquals(normaltxn.dialect, TEST_DIALECT)
        self.assertEquals(normaltxn.commandBlock().dialect, TEST_DIALECT)
        self.pauseHolders()
        extra = []
        extra.append(self.createTransaction())
        waitingtxn = self.createTransaction()
        self.assertEquals(waitingtxn.dialect, TEST_DIALECT)
        self.flushHolders()
        self.pool.stopService()
        notxn = self.createTransaction()
        self.assertEquals(notxn.dialect, TEST_DIALECT)


    def test_reConnectWhenFirstExecFails(self):
        """
        Generally speaking, DB-API 2.0 adapters do not provide information
        about the cause of a failed C{execute} method; they definitely don't
        provide it in a way which can be identified as related to the syntax of
        the query, the state of the database itself, the state of the
        connection, etc.

        Therefore the best general heuristic for whether the connection to the
        database has been lost and needs to be re-established is to catch
        exceptions which are raised by the I{first} statement executed in a
        transaction.
        """
        # Allow C{connect} to succeed.  This should behave basically the same
        # whether connect() happened to succeed in some previous transaction
        # and it's recycling the underlying transaction, or connect() just
        # succeeded.  Either way you just have a _SingleTxn wrapping a
        # _ConnectedTxn.
        txn = self.createTransaction()
        self.assertEquals(len(self.factory.connections), 1,
                          "Sanity check failed.")

        class CustomExecuteFailed(Exception):
            """
            Custom "execute-failed" exception.
            """

        self.factory.connections[0].executeWillFail(CustomExecuteFailed)
        results = self.resultOf(txn.execSQL("hello, world!"))
        [[[_ignore_counter, echo]]] = results
        self.assertEquals("hello, world!", echo)

        # Two execution attempts should have been made, one on each connection.
        # The first failed with a RuntimeError, but that is deliberately
        # obscured, because then we tried again and it succeeded.
        self.assertEquals(
            len(self.factory.connections), 2,
            "No new connection opened."
        )
        self.assertEquals(self.factory.connections[0].executions, 1)
        self.assertEquals(self.factory.connections[1].executions, 1)
        self.assertEquals(self.factory.connections[0].closed, True)
        self.assertEquals(self.factory.connections[1].closed, False)

        # Nevertheless, since there is currently no classification of "safe"
        # errors, we should probably log these messages when they occur.
        self.assertEquals(len(self.flushLoggedErrors(CustomExecuteFailed)), 1)


    def test_reConnectWhenFirstExecOnExistingConnectionFails(
        self, moreFailureSetup=lambda factory: None
    ):
        """
        Another situation that might arise is that a connection will be
        successfully connected, executed and recycled into the connection pool;
        then, the database server will shut down and the connections will die,
        but we will be none the wiser until we try to use them.
        """
        txn = self.createTransaction()
        moreFailureSetup(self.factory)
        self.assertEquals(
            len(self.factory.connections), 1, "Sanity check failed."
        )
        results = self.resultOf(txn.execSQL("hello, world!"))
        txn.commit()
        [[[_ignore_counter, echo]]] = results
        self.assertEquals("hello, world!", echo)
        txn2 = self.createTransaction()
        self.assertEquals(
            len(self.factory.connections), 1, "Sanity check failed."
        )

        class CustomExecFail(Exception):
            """
            Custom C{execute()} failure.
            """

        self.factory.connections[0].executeWillFail(CustomExecFail)
        results = self.resultOf(txn2.execSQL("second try!"))
        txn2.commit()
        [[[_ignore_counter, echo]]] = results
        self.assertEquals("second try!", echo)
        self.assertEquals(len(self.flushLoggedErrors(CustomExecFail)), 1)


    def test_closeExceptionDoesntHinderReconnection(self):
        """
        In some database bindings, if the server closes the connection,
        C{close()} will fail.  If C{close} fails, there's not much that could
        mean except that the connection is already closed, so similar to the
        condition described in
        L{test_reConnectWhenFirstExecOnExistingConnectionFails}, the
        failure should be logged, but transparent to application code.
        """

        class BindingSpecificException(Exception):
            """
            Exception that's a placeholder for something that a database
            binding might raise.
            """

        def alsoFailClose(factory):
            factory.childCloseWillFail(BindingSpecificException())

        t = self.test_reConnectWhenFirstExecOnExistingConnectionFails(
            alsoFailClose
        )
        errors = self.flushLoggedErrors(BindingSpecificException)
        self.assertEquals(len(errors), 1)
        return t


    def test_preCommitSuccess(self):
        """
        Callables passed to L{IAsyncTransaction.preCommit} will be invoked upon
        commit.
        """
        txn = self.createTransaction()

        def simple():
            simple.done = True

        simple.done = False
        txn.preCommit(simple)
        self.assertEquals(simple.done, False)
        result = self.resultOf(txn.commit())
        self.assertEquals(len(result), 1)
        self.assertEquals(simple.done, True)


    def test_deferPreCommit(self):
        """
        If callables passed to L{IAsyncTransaction.preCommit} return
        L{Deferred}s, they will defer the actual commit operation until it has
        fired.
        """
        txn = self.createTransaction()
        d = Deferred()

        def wait():
            wait.started = True

            def executed(it):
                wait.sqlResult = it

            # To make sure the _underlying_ commit operation was Deferred, we
            # have to execute some SQL to make sure it happens.
            d.addCallback(lambda ignored: txn.execSQL("some test sql"))
            d.addCallback(executed)
            return d

        wait.started = False
        wait.sqlResult = None
        txn.preCommit(wait)
        result = self.resultOf(txn.commit())
        self.flushHolders()
        self.assertEquals(wait.started, True)
        self.assertEquals(wait.sqlResult, None)
        self.assertEquals(result, [])
        d.callback(None)
        # allow network I/O for pooled / networked implementation; there should
        # be the commit message now.
        self.flushHolders()
        self.assertEquals(len(result), 1)
        self.assertEquals(wait.sqlResult, [[1, "some test sql"]])


    def test_failPreCommit(self):
        """
        If callables passed to L{IAsyncTransaction.preCommit} raise an
        exception or return a Failure, subsequent callables will not be run,
        and the transaction will be aborted.
        """
        def test(flawedCallable, exc):
            # Set up.
            test.committed = False
            test.aborted = False
            # Create transaction and add monitoring hooks.
            txn = self.createTransaction()

            def didCommit():
                test.committed = True

            def didAbort():
                test.aborted = True

            txn.postCommit(didCommit)
            txn.postAbort(didAbort)
            txn.preCommit(flawedCallable)
            result = self.resultOf(txn.commit())
            self.flushHolders()
            self.assertResultList(result, Failure(exc()))
            self.assertEquals(test.committed, False)
            self.assertEquals(test.aborted, True)

        def failer():
            return fail(ZeroDivisionError())

        def raiser():
            raise EOFError()

        test(failer, ZeroDivisionError)
        test(raiser, EOFError)


    def test_noOpCommitDoesntHinderReconnection(self):
        """
        Until you've executed a query or performed a statement on an ADBAPI
        connection, the connection is semantically idle (between transactions).
        A .commit() or .rollback() followed immediately by a .commit() is
        therefore pointless, and can be ignored.  Furthermore, actually
        executing the commit and propagating a possible connection-oriented
        error causes clients to see errors, when, if those clients had actually
        executed any statements, the connection would have been recycled and
        the statement transparently re-executed by the logic tested by
        L{test_reConnectWhenFirstExecFails}.
        """
        txn = self.createTransaction()
        self.factory.commitFail = True
        self.factory.rollbackFail = True
        [x] = self.resultOf(txn.commit())

        # No statements have been executed, so C{commit} will *not* be
        # executed.
        self.assertEquals(self.factory.commitFail, True)
        self.assertIdentical(x, None)
        self.assertEquals(len(self.pool._free), 1)
        self.assertEquals(self.pool._finishing, [])
        self.assertEquals(len(self.factory.connections), 1)
        self.assertEquals(self.factory.connections[0].closed, False)


    def test_reConnectWhenSecondExecFailsThenFirstExecFails(self):
        """
        Other connection-oriented errors might raise exceptions if they occur
        in the middle of a transaction, but that should cause the error to be
        caught, the transaction to be aborted, and the (closed) connection to
        be recycled, where the next transaction that attempts to do anything
        with it will encounter the error immediately and discover it needs to
        be recycled.

        It would be better if this behavior were invisible, but that could only
        be accomplished with more precise database exceptions.  We may come up
        with support in the future for more precisely identifying exceptions,
        but I{unknown} exceptions should continue to be treated in this manner,
        relaying the exception back to application code but attempting a
        re-connection on the next try.
        """
        txn = self.createTransaction()
        [[[_ignore_counter, _ignore_echo]]] = self.resultOf(txn.execSQL("hello, world!", []))
        self.factory.connections[0].executeWillFail(ZeroDivisionError)
        [f] = self.resultOf(txn.execSQL("divide by zero", []))
        f.trap(self.translateError(ZeroDivisionError))
        self.assertEquals(self.factory.connections[0].executions, 2)

        # Reconnection should work exactly as before.
        self.assertEquals(self.factory.connections[0].closed, False)

        # Application code has to roll back its transaction at this point,
        # since it failed (and we don't necessarily know why it failed: not
        # enough information).
        self.resultOf(txn.abort())
        self.factory.connections[0].executions = 0  # re-set for next test
        self.assertEquals(len(self.factory.connections), 1)
        self.test_reConnectWhenFirstExecFails()


    def test_disconnectOnFailedRollback(self):
        """
        When C{rollback} fails for any reason on a connection object, then we
        don't know what state it's in.  Most likely, it's already been
        disconnected, so the connection should be closed and the transaction
        de-pooled instead of recycled.

        Also, a new connection will immediately be established to keep the pool
        size the same.
        """
        txn = self.createTransaction()
        results = self.resultOf(txn.execSQL("maybe change something!"))
        [[[_ignore_counter, echo]]] = results
        self.assertEquals("maybe change something!", echo)
        self.factory.rollbackFail = True
        [x] = self.resultOf(txn.abort())

        # Abort does not propagate the error on, the transaction merely gets
        # disposed of.
        self.assertIdentical(x, None)
        self.assertEquals(len(self.pool._free), 1)
        self.assertEquals(self.pool._finishing, [])
        self.assertEquals(len(self.factory.connections), 2)
        self.assertEquals(self.factory.connections[0].closed, True)
        self.assertEquals(self.factory.connections[1].closed, False)
        self.assertEquals(len(self.flushLoggedErrors(RollbackFail)), 1)


    def test_exceptionPropagatesFailedCommit(self):
        """
        A failed C{rollback} is fine (the premature death of the connection
        without C{commit} means that the changes are surely gone), but a failed
        C{commit} has to be relayed to client code, since that actually means
        some changes didn't hit the database.
        """
        txn = self.createTransaction()
        self.factory.commitFail = True
        results = self.resultOf(txn.execSQL("maybe change something!"))
        [[[_ignore_counter, echo]]] = results
        self.assertEquals("maybe change something!", echo)
        [x] = self.resultOf(txn.commit())
        x.trap(self.translateError(CommitFail))

        self.assertEquals(len(self.pool._free), 1)
        self.assertEquals(self.pool._finishing, [])
        self.assertEquals(len(self.factory.connections), 2)
        self.assertEquals(self.factory.connections[0].closed, True)
        self.assertEquals(self.factory.connections[1].closed, False)


    def test_commandBlock(self):
        """
        L{IAsyncTransaction.commandBlock} returns an L{IAsyncTransaction}
        provider which ensures that a block of commands are executed together.
        """
        txn = self.createTransaction()
        a = self.resultOf(txn.execSQL("a"))
        cb = txn.commandBlock()
        verifyObject(ICommandBlock, cb)
        b = self.resultOf(cb.execSQL("b"))
        d = self.resultOf(txn.execSQL("d"))
        c = self.resultOf(cb.execSQL("c"))
        cb.end()
        e = self.resultOf(txn.execSQL("e"))

        self.assertEquals(
            self.factory.connections[0].cursors[0].allExecutions,
            [("a", []), ("b", []), ("c", []), ("d", []), ("e", [])]
        )
        self.assertEquals(len(a), 1)
        self.assertEquals(len(b), 1)
        self.assertEquals(len(c), 1)
        self.assertEquals(len(d), 1)
        self.assertEquals(len(e), 1)


    def test_commandBlockWithLatency(self):
        """
        A block returned by L{IAsyncTransaction.commandBlock} won't start
        executing until all SQL statements scheduled before it have completed.
        """
        self.pauseHolders()
        txn = self.createTransaction()
        a = self.resultOf(txn.execSQL("a"))
        b = self.resultOf(txn.execSQL("b"))
        cb = txn.commandBlock()
        c = self.resultOf(cb.execSQL("c"))
        d = self.resultOf(cb.execSQL("d"))
        e = self.resultOf(txn.execSQL("e"))
        cb.end()
        self.flushHolders()

        self.assertEquals(
            self.factory.connections[0].cursors[0].allExecutions,
            [("a", []), ("b", []), ("c", []), ("d", []), ("e", [])]
        )

        self.assertEquals(len(a), 1)
        self.assertEquals(len(b), 1)
        self.assertEquals(len(c), 1)
        self.assertEquals(len(d), 1)
        self.assertEquals(len(e), 1)


    def test_twoCommandBlocks(self, flush=lambda: None):
        """
        When execution of one command block is complete, it will proceed to the
        next queued block, then to regular SQL executed on the transaction.
        """
        txn = self.createTransaction()
        cb1 = txn.commandBlock()
        cb2 = txn.commandBlock()
        txn.execSQL("e")
        cb1.execSQL("a")
        cb2.execSQL("c")
        cb1.execSQL("b")
        cb2.execSQL("d")
        cb2.end()
        cb1.end()
        flush()
        self.flushHolders()
        self.assertEquals(
            self.factory.connections[0].cursors[0].allExecutions,
            [("a", []), ("b", []), ("c", []), ("d", []), ("e", [])]
        )


    def test_twoCommandBlocksLatently(self):
        """
        Same as L{test_twoCommandBlocks}, but with slower callbacks.
        """
        self.pauseHolders()
        self.test_twoCommandBlocks(self.flushHolders)


    def test_commandBlockEndTwice(self):
        """
        L{CommandBlock.end} will raise L{AlreadyFinishedError} when called more
        than once.
        """
        txn = self.createTransaction()
        block = txn.commandBlock()
        block.end()
        self.assertRaises(AlreadyFinishedError, block.end)


    def test_commandBlockDelaysCommit(self):
        """
        Some command blocks need to run asynchronously, without the overall
        transaction-managing code knowing how far they've progressed.
        Therefore when you call {IAsyncTransaction.commit}(), it should not
        actually take effect if there are any pending command blocks.
        """
        txn = self.createTransaction()
        block = txn.commandBlock()
        commitResult = self.resultOf(txn.commit())
        self.resultOf(block.execSQL("in block"))
        self.assertEquals(commitResult, [])
        self.assertEquals(
            self.factory.connections[0].cursors[0].allExecutions,
            [("in block", [])]
        )
        block.end()
        self.flushHolders()
        self.assertEquals(commitResult, [None])


    def test_commandBlockDoesntDelayAbort(self):
        """
        A L{CommandBlock} can't possibly have anything interesting to say about
        a transaction that gets rolled back, so C{abort} applies immediately;
        all outstanding C{execSQL}s will fail immediately, on both command
        blocks and on the transaction itself.
        """
        txn = self.createTransaction()
        block = txn.commandBlock()
        block2 = txn.commandBlock()
        abortResult = self.resultOf(txn.abort())
        self.assertEquals(abortResult, [None])
        self.assertRaises(AlreadyFinishedError, block2.execSQL, "bar")
        self.assertRaises(AlreadyFinishedError, block.execSQL, "foo")
        self.assertRaises(AlreadyFinishedError, txn.execSQL, "baz")
        self.assertEquals(
            self.factory.connections[0].cursors[0].allExecutions, []
        )
        # end() should _not_ raise an exception, because this is the sort of
        # thing that might be around a try/finally or try/except; it's just
        # putting the commandBlock itself into a state consistent with the
        # transaction.
        block.end()
        block2.end()


    def test_endedBlockDoesntExecuteMoreSQL(self):
        """
        Attempting to execute SQL on a L{CommandBlock} which has had C{end}
        called on it will result in an L{AlreadyFinishedError}.
        """
        txn = self.createTransaction()
        block = txn.commandBlock()
        block.end()
        self.assertRaises(AlreadyFinishedError, block.execSQL, "hello")
        self.assertEquals(
            self.factory.connections[0].cursors[0].allExecutions, []
        )


    def test_commandBlockAfterCommitRaises(self):
        """
        Once an L{IAsyncTransaction} has been committed, L{commandBlock} raises
        an exception.
        """
        txn = self.createTransaction()
        txn.commit()
        self.assertRaises(AlreadyFinishedError, txn.commandBlock)


    def test_commandBlockAfterAbortRaises(self):
        """
        Once an L{IAsyncTransaction} has been committed, L{commandBlock} raises
        an exception.
        """
        txn = self.createTransaction()
        self.resultOf(txn.abort())
        self.assertRaises(AlreadyFinishedError, txn.commandBlock)


    def test_raiseOnZeroRowCount(self):
        """
        L{IAsyncTransaction.execSQL} will return a L{Deferred} failing with the
        exception passed as its raiseOnZeroRowCount argument if the underlying
        query returns no rows.
        """
        self.factory.hasResults = False
        txn = self.createTransaction()
        f = self.resultOf(
            txn.execSQL("hello", raiseOnZeroRowCount=ZeroDivisionError)
        )[0]
        self.assertRaises(ZeroDivisionError, f.raiseException)
        txn.commit()


    def test_raiseOnZeroRowCountWithUnreliableRowCount(self):
        """
        As it turns out, some databases can't reliably tell you how many rows
        they're going to fetch via the C{rowcount} attribute before the rows
        have actually been fetched, so the C{raiseOnZeroRowCount} will I{not}
        raise an exception if C{rowcount} is zero but C{description} and
        C{fetchall} indicates the presence of some rows.
        """
        self.factory.hasResults = True
        self.factory.shouldUpdateRowcount = False
        txn = self.createTransaction()
        r = self.resultOf(
            txn.execSQL("some-rows", raiseOnZeroRowCount=RuntimeError)
        )
        [[[_ignore_counter, echo]]] = r
        self.assertEquals(echo, "some-rows")



class IOPump(object):
    """
    Connect a client and a server.

    @ivar client: a client protocol

    @ivar server: a server protocol
    """

    def __init__(self, client, server):
        self.client = client
        self.server = server
        self.clientTransport = StringTransport()
        self.serverTransport = StringTransport()
        self.client.makeConnection(self.clientTransport)
        self.server.makeConnection(self.serverTransport)
        self.c2s = [self.clientTransport, self.server]
        self.s2c = [self.serverTransport, self.client]


    def moveData(self, (outTransport, inProtocol)):
        """
        Move data from a L{StringTransport} to an L{IProtocol}.

        @return: C{True} if any data was moved, C{False} if no data was moved.
        """
        data = outTransport.io.getvalue()
        outTransport.io.seek(0)
        outTransport.io.truncate()
        if data:
            inProtocol.dataReceived(data)
            return True
        else:
            return False


    def pump(self):
        """
        Deliver all input from the client to the server, then from the server
        to the client.
        """
        a = self.moveData(self.c2s)
        b = self.moveData(self.s2c)
        return a or b


    def flush(self, maxTurns=100):
        """
        Continue pumping until no more data is flowing.
        """
        turns = 0
        while self.pump():
            turns += 1
            if turns > maxTurns:
                raise RuntimeError("Ran too long!")



class NetworkedPoolHelper(ConnectionPoolHelper):
    """
    An extension of L{ConnectionPoolHelper} that can set up a
    L{ConnectionPoolClient} and L{ConnectionPoolConnection} attached to each
    other.
    """

    def setUp(self):
        """
        Do the same setup from L{ConnectionPoolBase}, but also establish a
        loopback connection between a L{ConnectionPoolConnection} and a
        L{ConnectionPoolClient}.
        """
        super(NetworkedPoolHelper, self).setUp()
        self.pump = IOPump(
            ConnectionPoolClient(
                dialect=self.dialect,
                paramstyle=self.paramstyle
            ),
            ConnectionPoolConnection(self.pool)
        )


    def flushHolders(self):
        """
        In addition to flushing the L{ThreadHolder} stubs, also flush any
        pending network I/O.
        """
        self.pump.flush()
        super(NetworkedPoolHelper, self).flushHolders()
        self.pump.flush()


    def createTransaction(self):
        txn = self.pump.client.newTransaction()
        self.pump.flush()
        return txn


    def translateError(self, err):
        """
        All errors raised locally will unfortunately be translated into
        UnknownRemoteError, since AMP requires specific enumeration of all of
        them.  Flush the locally logged error of the given type and return
        L{UnknownRemoteError}.
        """
        if err in Commit.errors:
            return err
        self.flushLoggedErrors(err)
        return FailsafeException


    def resultOf(self, it):
        result = resultOf(it)
        self.pump.flush()
        return result



class NetworkedConnectionPoolTests(NetworkedPoolHelper, ConnectionPoolTests):
    """
    Tests for L{ConnectionPoolConnection} and L{ConnectionPoolClient}
    interacting with each other.
    """

    def setParamstyle(self, paramstyle):
        """
        Change the paramstyle on both the pool and the client.
        """
        super(NetworkedConnectionPoolTests, self).setParamstyle(paramstyle)
        self.pump.client.paramstyle = paramstyle


    def setDialect(self, dialect):
        """
        Change the dialect on both the pool and the client.
        """
        super(NetworkedConnectionPoolTests, self).setDialect(dialect)
        self.pump.client.dialect = dialect


    def test_newTransaction(self):
        """
        L{ConnectionPoolClient.newTransaction} returns a provider of
        L{IAsyncTransaction}, and creates a new transaction on the server side.
        """
        txn = self.pump.client.newTransaction()
        verifyObject(IAsyncTransaction, txn)
        self.pump.flush()
        self.assertEquals(len(self.factory.connections), 1)



class HookableOperationTests(TestCase):
    """
    Tests for L{_HookableOperation}.
    """

    @inlineCallbacks
    def test_clearPreventsSubsequentAddHook(self):
        """
        After clear() or runHooks() are called, subsequent calls to addHook()
        are NO-OPs.
        """
        def hook():
            return succeed(None)

        hookOp = _HookableOperation()
        hookOp.addHook(hook)
        self.assertEquals(len(hookOp._hooks), 1)
        hookOp.clear()
        self.assertEquals(hookOp._hooks, None)

        hookOp = _HookableOperation()
        hookOp.addHook(hook)
        yield hookOp.runHooks()
        self.assertEquals(hookOp._hooks, None)
        hookOp.addHook(hook)
        self.assertEquals(hookOp._hooks, None)



##
# Copyright (c) 2010-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Decorators.
"""

__all__ = [
    "memoizedKey",
]


from inspect import getargspec

from twisted.internet.defer import Deferred, succeed

class Memoizable(object):
    """
    A class that stores itself in the memo dictionary.
    """

    def memoMe(self, key, memo):
        """
        Add this object to the memo dictionary in whatever fashion is appropriate.

        @param key: key used for lookup
        @type key: C{object} (typically C{str} or C{int})
        @param memo: the dict to store to
        @type memo: C{dict}
        """
        raise NotImplementedError



def memoizedKey(keyArgument, memoAttribute, deferredResult=True):
    """
    Decorator which memoizes the result of a method on that method's instance. If the instance is derived from
    class Memoizable, then the memoMe method is used to store the result, otherwise it is stored directly in
    the dict.

    @param keyArgument: The name of the "key" argument.
    @type keyArgument: C{str}

    @param memoAttribute: The name of the attribute on the instance which
        should be used for memoizing the result of this method; the attribute
        itself must be a dictionary.  Alternately, if the specified argument is
        callable, it is a callable that takes the arguments passed to the
        decorated method and returns the memo dictionaries.
    @type memoAttribute: C{str} or C{callable}

    @param deferredResult: Whether the result must be a deferred.
    """
    def getarg(argname, argspec, args, kw):
        """
        Get an argument from some arguments.

        @param argname: The name of the argument to retrieve.

        @param argspec: The result of L{inspect.getargspec}.

        @param args: positional arguments passed to the function specified by
            argspec.

        @param kw: keyword arguments passed to the function specified by
            argspec.

        @return: The value of the argument named by 'argname'.
        """
        argnames = argspec[0]
        try:
            argpos = argnames.index(argname)
        except ValueError:
            argpos = None
        if argpos is not None:
            if len(args) > argpos:
                return args[argpos]
        if argname in kw:
            return kw[argname]
        else:
            raise TypeError("could not find key argument %r in %r/%r (%r)" % (
                argname, args, kw, argpos
            ))


    def decorate(thunk):
        # cheater move to try to get the right argspec from inlineCallbacks.
        # This could probably be more robust, but the 'cell_contents' thing
        # probably can't (that's the only real reference to the underlying
        # function).
        if thunk.func_code.co_name == "unwindGenerator":
            specTarget = thunk.func_closure[0].cell_contents
        else:
            specTarget = thunk
        spec = getargspec(specTarget)

        def outer(*a, **kw):
            self = a[0]
            if callable(memoAttribute):
                memo = memoAttribute(*a, **kw)
            else:
                memo = getattr(self, memoAttribute)
            key = getarg(keyArgument, spec, a, kw)
            if key in memo:
                memoed = memo[key]
                if deferredResult:
                    return succeed(memoed)
                else:
                    return memoed
            result = thunk(*a, **kw)

            if isinstance(result, Deferred):
                def memoResult(finalResult):
                    if isinstance(finalResult, Memoizable):
                        finalResult.memoMe(key, memo)
                    elif finalResult is not None:
                        memo[key] = finalResult
                    return finalResult
                result.addCallback(memoResult)
            elif result is not None:
                memo[key] = result
            return result

        return outer

    return decorate
